from configparser import ConfigParser
from typing import Type, List
from sqlalchemy import func
from sqlalchemy.orm import Session
from gnatwriter.controllers import BaseController
from gnatwriter.models import User, Activity


class ActivityController(BaseController):
    """Activity controller encapsulates activity management functionality

    Attributes
    ----------
    _instance : ActivityController
        The instance of the activity controller
    _config : configParser
        The path to the configuration file
    _session : Session
        The database session object
    _owner : Type[User]
        The current user of the controller

    Methods
    -------
    get_activity_by_id(activity_id: int)
        Get an activity by id
    get_activities()
        Get all activities associated with a user, sorted by created date with most recent first
    get_activities_page(user_id: int, page: int, per_page: int)
        Get a single page of activities associated with a user, sorted by created date with most recent first
    get_activity_count()
        Get activity count associated with a user
    search_activities(search: str)
        Search for activities by summary
    search_activities_page(search: str, page: int, per_page: int)
        Search for activities by summary and get a single page of activities associated with a user, sorted by created
        date with most recent first
    """

    def __init__(
            self, config: ConfigParser, session: Session, owner: Type[User]
    ):
        """Initialize the class"""

        super().__init__(config, session, owner)

    def get_activity_by_id(self, activity_id: int) -> Type[Activity] | None:
        """Get an activity by id

        Parameters
        ----------
        activity_id : int
            The id of the activity to get

        Returns
        -------
        Activity | None
            The activity object or None if not found
        """

        with self._session as session:

            activity = session.query(Activity).filter(
                Activity.id == activity_id,
                Activity.user_id == self._owner.id
            ).first()

            return activity if activity else None

    def get_activities(self) -> List[Type[Activity]]:
        """Get all activities associated with a user, sorted by created date with most recent first

        Returns
        -------
        list
            A list of activity objects
        """

        with self._session as session:

            return session.query(Activity).filter(
                Activity.user_id == self._owner.id
            ).order_by(Activity.created.desc()).all()

    def get_activities_page(
        self, page: int, per_page: int
    ) -> List[Type[Activity]]:
        """Get a single page of activities associated with a user, sorted by created date with most recent first

        Parameters
        ----------
        page : int
            The page number
        per_page : int
            The number of rows per page

        Returns
        -------
        list
            A list of activity objects
        """

        with self._session as session:

            offset = (page - 1) * per_page

            return session.query(Activity).filter(
                Activity.user_id == self._owner.id
            ).order_by(
                Activity.created.desc()
            ).offset(offset).limit(per_page).all()

    def get_activity_count(self) -> int:
        """Get activity count associated with a user

        Returns
        -------
        int
            The number of activities
        """

        with self._session as session:

            return session.query(func.count(Activity.id)).filter(
                Activity.user_id == self._owner.id
            ).scalar()

    def search_activities(self, search: str) -> List[Type[Activity]]:
        """Search for activities by summary

        Parameters
        ----------
        search : str
            The search string

        Returns
        -------
        list
            A list of activity objects
        """

        with self._session as session:

            return session.query(Activity).filter(
                Activity.summary.like(f'%{search}%'),
                Activity.user_id == self._owner.id
            ).all()

    def search_activities_page(
        self, search: str, page: int, per_page: int
    ) -> List[Type[Activity]]:
        """Search for activities by summary and get a single page of activities associated with a user

        Parameters
        ----------
        search : str
            The search string
        page : int
            The page number
        per_page : int
            The number of rows per page

        Returns
        -------
        list
            A list of activity objects
        """

        with self._session as session:

            offset = (page - 1) * per_page

            return session.query(Activity).filter(
                Activity.summary.like(f'%{search}%'),
                Activity.user_id == self._owner.id
            ).order_by(
                Activity.created.desc()
            ).offset(offset).limit(per_page).all()



from datetime import datetime
from sqlalchemy import Integer, ForeignKey, DateTime
from sqlalchemy.orm import Mapped, mapped_column, relationship
from gnatwriter.models import User, Author, Story, Base


class AuthorStory(Base):
    """The AuthorStory class represents the relationship between an Author and a Story.

    Attributes
    ----------
        author_id: int
            The author's id
        story_id: int
            The story's id
        user_id: int
            The id of the owner of this entry
        created: str
            The creation datetime of the link between the Author and the Story
        user: User
            The user who owns this entry
        author: Author
            The author name assigned to the story
        story: Story
            The story that the author (user) wrote

    Methods
    -------
        __repr__()
            Returns a string representation of the relationship
        __str__()
            Returns a string representation of the relationship
        serialize()
            Returns a dictionary representation of the relationship
        unserialize(data: dict)
            Updates the relationship's attributes with the values from the dictionary
    """

    __tablename__ = 'authors_stories'
    author_id: Mapped[int] = mapped_column(Integer, ForeignKey('authors.id'), primary_key=True)
    story_id: Mapped[int] = mapped_column(Integer, ForeignKey('stories.id'), primary_key=True)
    user_id: Mapped[int] = mapped_column(Integer, ForeignKey('users.id'))
    created: Mapped[str] = mapped_column(DateTime, default=str(datetime.now()))
    user: Mapped["User"] = relationship("User")
    author: Mapped["Author"] = relationship(
        "Author", back_populates="stories", lazy="joined"
    )
    story: Mapped["Story"] = relationship(
        "Story", back_populates="authors"
    )

    def __repr__(self):
        """Returns a string representation of the relationship.

        Returns
        -------
        str
            A string representation of the relationship
        """

        return f'<AuthorStory {self.author.name!r} - {self.story.title!r}>'

    def __str__(self):
        """Returns a string representation of the relationship.

        Returns
        -------
        str
            A string representation of the relationship
        """

        return f'AuthorStory: {self.author.name} - {self.story.title}'

    def serialize(self) -> dict:
        """Returns a dictionary representation of the relationship.

        Returns
        -------
        dict
            A dictionary representation of the relationship
        """

        return {
            'author_id': self.author_id,
            'story_id': self.story_id,
            'user_id': self.user_id,
            'created': str(self.created),
        }

    def unserialize(self, data: dict) -> "AuthorStory":
        """Updates the relationship's attributes with the values from the dictionary.

        Returns
        -------
        AuthorStory
            The updated relationship
        """

        self.author_id = data.get('author_id', self.author_id)
        self.story_id = data.get('story_id', self.story_id)
        self.user_id = data.get('user_id', self.user_id)
        self.created = data.get('created', self.created)

        return self



from configparser import ConfigParser
from datetime import datetime
from typing import Optional, List
from sqlalchemy import Integer, ForeignKey, String, Date, DateTime
from sqlalchemy.orm import Mapped, mapped_column, relationship, validates
from gnatwriter.models import User, Story, BibliographyAuthor, Base


class Bibliography(Base):
    """The Bibliography class represents a reference to a story.

    Attributes
    ----------
        id: int
            The Bibliography ID
        user_id: int
            The ID of the user who created this Bibliography entry
        story_id: int
            The Story ID that this Bibliography references
        title: str
            The referenced work's title
        pages: str
            The pertinent pages of the referenced work
        publisher: str
            The referenced work's publisher
        publication_date: str
            The referenced work's publication date in date form: yyy-mm-dd
        editor: str
            The referenced work's editor
        created: str
            The referenced work's creation date in datetime form: yyy-mm-dd hh:mm:ss
        modified: str
            The referenced work's last modification date in datetime form: yyy-mm-dd hh:mm:ss
        user: User
            The user who created this Bibliography entry
        story: Story
            The Story to which this Bibliography entry belongs
        authors: List[BibliographyAuthor]
            The author(s) of the referenced work

    Methods
    -------
        __repr__()
            Returns a string representation of the reference
        __str__()
            Returns a string representation of the reference
        serialize()
            Returns a dictionary representation of the reference
        unserialize(data: dict)
            Updates the reference's attributes with the values from the dictionary
        validate_title(title: str)
            Validates the title's length
        validate_pages(pages: str)
            Validates the pages' length
        validate_publication_date(publication_date: str)
            Validates the publication date's format
    """

    __tablename__ = 'bibliographies'
    id: Mapped[int] = mapped_column(Integer, primary_key=True)
    user_id: Mapped[int] = mapped_column(Integer, ForeignKey('users.id'))
    story_id: Mapped[int] = mapped_column(Integer, ForeignKey('stories.id'))
    title: Mapped[str] = mapped_column(String(250), nullable=True)
    pages: Mapped[str] = mapped_column(String(50), nullable=True)
    publisher: Mapped[str] = mapped_column(String(100), nullable=True)
    publication_date: Mapped[str] = mapped_column(Date, nullable=True)
    editor: Mapped[str] = mapped_column(String(100), nullable=True)
    created: Mapped[str] = mapped_column(DateTime, default=str(datetime.now()))
    modified: Mapped[str] = mapped_column(
        DateTime, default=str(datetime.now()), onupdate=str(datetime.now())
    )
    user: Mapped["User"] = relationship("User")
    story: Mapped["Story"] = relationship("Story", back_populates="references")
    authors: Mapped[Optional[List["BibliographyAuthor"]]] = relationship(
        "BibliographyAuthor", back_populates="reference",
        lazy="joined", cascade="all, delete, delete-orphan"
    )

    def __repr__(self):
        """Returns a string representation of the reference.

        Returns
        -------
        str
            A string representation of the reference
        """

        return f'<Bibliography {self.title!r}>'

    def __str__(self):
        """Returns a string representation of the reference.

        Returns
        -------
        str
            A string representation of the reference
        """

        return f'Bibliography: {self.title}'

    def serialize(self) -> dict:
        """Returns a dictionary representation of the reference.

        Returns
        -------
        dict
            A dictionary representation of the reference
        """

        return {
            'id': self.id,
            'user_id': self.user_id,
            'story_id': self.story_id,
            'title': self.title,
            'pages': self.pages,
            'publication_date': str(self.publication_date),
            'created': str(self.created),
            'modified': str(self.modified),
        }

    def unserialize(self, data: dict) -> "Bibliography":
        """Updates the reference's attributes with the values from the dictionary.


        Returns
        -------
        Bibliography
            The updated reference
        """

        self.id = data.get('id', self.id)
        self.user_id = data.get('user_id', self.user_id)
        self.story_id = data.get('story_id', self.story_id)
        self.title = data.get('title', self.title)
        self.pages = data.get('pages', self.pages)
        self.publication_date = data.get('publication_date', self.publication_date)
        self.created = data.get('created', self.created)
        self.modified = data.get('modified', self.modified)

        return self

    @validates("title")
    def validate_title(self, key, title: str) -> str:
        """Validates the title's length.

        Returns
        -------
        str
            The validated title
        """

        if not title:
            raise ValueError("Reference title must not be empty.")

        if len(title) > 250:
            raise ValueError("Reference title must not be longer than 250 characters.")

        return title

    @validates("pages")
    def validate_pages(self, key, pages: str) -> str:
        """Validates the pages' length.

        Returns
        -------
        str
            The validated pages
        """

        if pages and len(pages) > 50:
            raise ValueError("Reference pages data must not be longer than 50 characters.")

        return pages

    @validates("publisher")
    def validate_publisher(self, key, publisher: str) -> str:
        """Validates the publisher's length.

        Parameters
        ----------
        publisher: str
            The reference's publisher

        Returns
        -------
        str
            The validated publisher
        """

        if publisher and len(publisher) > 100:
            raise ValueError("Reference publisher data must not be longer than 100 characters.")

        return publisher

    @validates("publication_date")
    def validate_publication_date(self, key, publication_date: str) -> str | None:
        """Validates the publication date's format.

        Returns
        -------
        str
            The validated publication date
        """

        config = ConfigParser()
        config.read("config.cfg")
        date_format = config.get("formats", "date")

        if publication_date is not None and bool(datetime.strptime(publication_date, date_format)) is False:
            raise ValueError("Reference publication date must be in the format 'YYYY-MM-DD'.")

        return publication_date

    @validates("editor")
    def validate_editor(self, key, editor: str) -> str:
        """Validates the editor's length.

        Parameters
        ----------
        editor: str
            The reference's editor

        Returns
        -------
        str
            The validated editor
        """

        if editor and len(editor) > 100:
            raise ValueError("Reference editor data must not be longer than 100 characters.")

        return editor



from configparser import ConfigParser
from datetime import datetime, date
from typing import Type, List
from sqlalchemy.orm import Session
from gnatwriter.controllers.BaseController import BaseController
from gnatwriter.models import User, Submission, Activity


class SubmissionController(BaseController):
    """Submission controller encapsulates submission management functionality

    Attributes
    ----------
    _instance : SubmissionController
        The instance of the submission controller
    _config: ConfigParser
        The configuration parser
    _owner : User
        The current user of the submission controller
    _session : Session
        The database session

    Methods
    -------
    create_submission(story_id: int, submitted_to: str, date_sent: str)
        Create a new submission
    update_submission(submission_id: int, submitted_to: str, date_sent: str, date_reply_received: str, \
                      date_published: str, date_paid: str, result: str, amount: float)
        Update a submission
    delete_submission_by_id(submission_id: int)
        Delete a submission
    get_all_submissions()
        Get all submissions associated with an owner
    get_all_submissions_page(page: int, per_page: int)
        Get a single page of submissions associated with an owner from the database
    get_submissions_by_story_id(story_id: int)
        Get all submissions associated with a story
    get_submissions_page_by_story_id(story_id: int, page
        Get a single page of submissions associated with a story from the database
    """

    def __init__(
        self, config: ConfigParser, session: Session, owner: Type[User]
    ):
        """Initialize the class"""

        super().__init__(config, session, owner)

    def create_submission(
        self, story_id: int, submitted_to: str, date_sent: str = None,
        date_reply_received: str = None, date_published: str = None,
        date_paid: str = None, result: str = None, amount: float = None
    ) -> Submission:
        """Create a new submission

        Parameters
        ----------
        story_id : int
            The id of the story
        submitted_to : str
            The entity the story was submitted to
        date_sent : str
            The date the story was submitted, optional
        date_reply_received : str
            The date the reply was received
        date_published : str
            The date the story was published
        date_paid : str
            The date on which the story was paid
        result : str
            The result of the submission
        amount : float
            The amount paid for the submission

        Returns
        -------
        Submission
            The new submission object
        """

        with (self._session as session):

            try:

                if date_sent:
                    date_sent = datetime.strptime(date_sent, '%Y-%m-%d').date()
                else:
                    date_sent = None

                if date_reply_received:
                    date_reply_received = datetime.strptime(date_reply_received, '%Y-%m-%d').date()
                else:
                    date_reply_received = None

                if date_published:
                    date_published = datetime.strptime(date_published, '%Y-%m-%d').date()
                else:
                    date_published = None

                if date_paid:
                    date_paid = datetime.strptime(date_paid, '%Y-%m-%d').date()
                else:
                    date_paid = None

                result_states = [
                    'Pending', 'Rewrite Requested', 'Ignored', 'Withdrawn',
                    'Rejected', 'Accepted'
                ]

                if result not in result_states:
                    raise ValueError('Invalid submission result.')

                created = datetime.now()
                modified = created

                submission = Submission(
                    user_id=self._owner.id, story_id=story_id,
                    submitted_to=submitted_to, date_sent=date_sent,
                    date_reply_received=date_reply_received,
                    date_published=date_published, date_paid=date_paid,
                    result=result, amount=amount, created=created,
                    modified=modified
                )

                activity = Activity(
                    user_id=self._owner.id, summary=f'Submission \
                    {submission.id} created by {self._owner.username}',
                    created=datetime.now()
                )

                session.add(submission)
                session.add(activity)

            except Exception as e:
                session.rollback()
                raise e

            else:
                session.commit()
                return submission

    def update_submission(
        self, submission_id: int, submitted_to: str, date_sent: str = None,
        date_reply_received: str = None, date_published: str = None,
        date_paid: str = None, result: str = None, amount: float = None
    ) -> Type[Submission]:
        """Update a submission

        Parameters
        ----------
        submission_id : int
            The id of the submission
        submitted_to : str
            The entity the story was submitted to
        date_sent : str
            The date the story was submitted
        date_reply_received : str
            The date the reply was received
        date_published : str
            The date the story was published
        date_paid : str
            The date on which the story was paid
        result : str
            The result of the submission
        amount : float
            The amount paid for the submission

        Returns
        -------
        Submission
            The updated submission object
        """

        with self._session as session:
            try:
                submission = session.query(Submission).filter(
                    Submission.id == submission_id,
                    Submission.user_id == self._owner.id
                ).first()

                if not submission:
                    raise ValueError('Submission not found.')

                submission.submitted_to = submitted_to

                if date_sent:
                    submission.date_sent = datetime.strptime(date_sent, '%Y-%m-%d').date()
                else:
                    submission.date_sent = None

                if date_reply_received:
                    submission.date_reply_received = datetime.strptime(date_reply_received, '%Y-%m-%d').date()
                else:
                    submission.date_reply_received = None

                if date_published:
                    submission.date_published = datetime.strptime(date_published, '%Y-%m-%d').date()
                else:
                    submission.date_published = None

                if date_paid:
                    submission.date_paid = datetime.strptime(date_paid, '%Y-%m-%d').date()
                else:
                    submission.date_paid = None

                result_states = [
                    'Pending', 'Rewrite Requested', 'Ignored', 'Withdrawn',
                    'Rejected', 'Accepted'
                ]

                if result not in result_states:
                    raise ValueError('Invalid submission result.')

                submission.result = result
                submission.amount = amount
                submission.modified = datetime.now()

                activity = Activity(
                    user_id=self._owner.id, summary=f'Submission \
                    {submission.id} updated by {self._owner.username}',
                    created=datetime.now()
                )

                session.add(activity)

            except Exception as e:
                session.rollback()
                raise e

            else:
                session.commit()
                return submission

    def delete_submission_by_id(self, submission_id: int) -> bool:
        """Delete a submission

        Parameters
        ----------
        submission_id : int
            The id of the submission

        Returns
        -------
        bool
            True on success
        """

        with self._session as session:
            try:
                submission = session.query(Submission).filter(
                    Submission.id == submission_id,
                    Submission.user_id == self._owner.id
                ).first()

                if not submission:
                    raise ValueError('Submission not found.')

                activity = Activity(
                    user_id=self._owner.id, summary=f'Submission \
                    {submission.id} deleted by {self._owner.username}',
                    created=datetime.now()
                )

                session.delete(submission)
                session.add(activity)

            except Exception as e:
                session.rollback()
                raise e

            else:
                session.commit()
                return True

    def get_submission_by_id(self, submission_id: int) -> Type[Submission]:
        """Get a submission by id

        Parameters
        ----------
        submission_id : int
            The id of the submission

        Returns
        -------
        Submission
            The submission object
        """

        with self._session as session:
            return session.query(Submission).filter(
                Submission.id == submission_id,
                Submission.user_id == self._owner.id
            ).first()

    def get_all_submissions(self) -> List[Type[Submission]]:
        """Get all submissions associated with an owner

        Returns
        -------
        list
            A list of submission objects
        """

        with self._session as session:

            return session.query(Submission).filter(
                Submission.user_id == self._owner.id
            ).all()

    def get_all_submissions_page(
        self, page: int, per_page: int
    ) -> List[Type[Submission]]:
        """Get a single page of submissions associated with an owner from the database

        Parameters
        ----------
        page : int
            The page number
        per_page : int
            The number of rows per page

        Returns
        -------
        list
            A list of submission objects
        """

        with self._session as session:

            offset = (page - 1) * per_page

            return session.query(Submission).filter(
                Submission.user_id == self._owner.id
            ).offset(offset).limit(per_page).all()

    def get_submissions_by_story_id(
        self, story_id: int
    ) -> List[Type[Submission]]:
        """Get all submissions associated with a story

        Parameters
        ----------
        story_id : int
            The id of the story

        Returns
        -------
        list
            A list of submission objects
        """

        with self._session as session:
            return session.query(Submission).filter(
                Submission.story_id == story_id,
                Submission.user_id == self._owner.id
            ).all()

    def get_submissions_page_by_story_id(
        self, story_id: int, page: int, per_page: int
    ) -> List[Type[Submission]]:
        """Get a single page of submissions associated with a story from the database

        Parameters
        ----------
        story_id : int
            The id of the story
        page : int
            The page number
        per_page : int
            The number of rows per page

        Returns
        -------
        list
            A list of submission objects
        """

        with self._session as session:

            offset = (page - 1) * per_page

            return session.query(Submission).filter(
                Submission.story_id == story_id,
                Submission.user_id == self._owner.id
            ).offset(offset).limit(per_page).all()



from datetime import datetime
from sqlalchemy import Integer, ForeignKey, DateTime
from sqlalchemy.orm import Mapped, mapped_column, relationship
from gnatwriter.models import User, Note, Scene, Base


class NoteScene(Base):
    """The NoteScene class represents the relationship between a note and a scene.

    Attributes
    ----------
        user_id: int
            The id of the owner of this entry
        note_id: int
            The note's id
        scene_id: int
            The scene's id
        created: str
            The creation datetime of the link between the Note and the Scene
        user: User
            The user who owns this entry
        note: Note
            The note that the scene belongs to
        scene: Scene
            The scene that the note has

    Methods
    -------
        __repr__()
            Returns a string representation of the relationship
        __str__()
            Returns a string representation of the relationship
        serialize()
            Returns a dictionary representation of the relationship
        unserialize(data: dict)
            Updates the relationship's attributes with the values from the dictionary
    """

    __tablename__ = 'notes_scenes'
    user_id: Mapped[int] = mapped_column(Integer, ForeignKey('users.id'))
    note_id: Mapped[int] = mapped_column(Integer, ForeignKey('notes.id'), primary_key=True)
    scene_id: Mapped[int] = mapped_column(Integer, ForeignKey('scenes.id'), primary_key=True)
    created: Mapped[str] = mapped_column(DateTime, default=str(datetime.now()))
    user: Mapped["User"] = relationship("User")
    note: Mapped["Note"] = relationship(
        "Note", back_populates="scenes", lazy="joined"
    )
    scene: Mapped["Scene"] = relationship(
        "Scene", back_populates="notes"
    )

    def __repr__(self):
        """Returns a string representation of the relationship.

        Returns
        -------
        str
            A string representation of the relationship
        """

        return f'<NoteScene {self.note.title!r} - {self.scene.title!r}>'

    def __str__(self):
        """Returns a string representation of the relationship.

        Returns
        -------
        str
            A string representation of the relationship
        """

        return f'{self.note.title} - {self.scene.title}'

    def serialize(self) -> dict:
        """Returns a dictionary representation of the relationship.

        Returns
        -------
        dict
            A dictionary representation of the relationship
        """

        return {
            'user_id': self.user_id,
            'note_id': self.note_id,
            'scene_id': self.scene_id,
            'created': str(self.created),
        }

    def unserialize(self, data: dict) -> "NoteScene":
        """Updates the relationship's attributes with the values from the dictionary.

        Parameters
        ----------
        data: dict
            The dictionary with the new values for the relationship

        Returns
        -------
        NoteScene
            The unserialized relationship
        """

        self.user_id = data.get('user_id', self.user_id)
        self.note_id = data.get('note_id', self.note_id)
        self.scene_id = data.get('scene_id', self.scene_id)
        self.created = data.get('created', self.created)

        return self



from datetime import datetime
from sqlalchemy import Integer, ForeignKey, Text, Date, String, Float, DateTime
from sqlalchemy.orm import Mapped, mapped_column, relationship, validates
from gnatwriter.models import SubmissionResultType, User, Story, Base


class Submission(Base):
    """The Submission class represents a submission of a story to some publisher or contest.

    Attributes
    ----------
        id: int
            The submission's id
        user_id: int
            The id of the owner of this entry
        story_id: int
            The story's id
        submitted_to: str
            The submission's recipient
        date_sent: str
            The submission's sent date
        date_reply_received: str
            The submission's reply received date
        date_published: str
            The submission's published date
        date_paid: str
            The submission's paid date
        result (SubmissionResultType):
            The submission's result
        amount: float
            The submission's amount
        created: str
            The creation datetime of the submission
        modified: str
            The last modification datetime of the submission
        user : User
            The user who owns this entry
        story: Story
            The story that the submission has

    Methods
    -------
        __repr__()
            Returns a string representation of the submission
        __str__()
            Returns a string representation of the submission
        serialize()
            Returns a dictionary representation of the submission
        unserialize(data: dict):
            Updates the submission's attributes with the values from the dictionary
        validate_submitted_to(submitted_to: str)
            Validates the submitted_to's length
        validate_date_sent(date_sent: str)
            Validates the date_sent's format
        validate_date_reply_received(date_reply_received: str)
            Validates the date_reply_received's format
        validate_date_published(date_published: str)
            Validates the date_published's format
        validate_date_paid(date_paid: str)
            Validates the date_paid's format
        validate_result(result: str)
            Validates the result's value
    """

    __tablename__ = 'submissions'
    id: Mapped[int] = mapped_column(Integer, primary_key=True)
    user_id: Mapped[int] = mapped_column(Integer, ForeignKey('users.id'))
    story_id: Mapped[int] = mapped_column(Integer, ForeignKey('stories.id'))
    submitted_to: Mapped[str] = mapped_column(Text, nullable=False)
    date_sent: Mapped[str] = mapped_column(Date, nullable=True)
    date_reply_received: Mapped[str] = mapped_column(Date, nullable=True)
    date_published: Mapped[str] = mapped_column(Date, nullable=True)
    date_paid: Mapped[str] = mapped_column(Date, nullable=True)
    result: Mapped[SubmissionResultType] = mapped_column(String(50), nullable=True)
    amount: Mapped[float] = mapped_column(Float, nullable=True)
    created: Mapped[str] = mapped_column(DateTime, default=str(datetime.now()))
    modified: Mapped[str] = mapped_column(DateTime, default=str(datetime.now()), onupdate=str(datetime.now()))
    user: Mapped["User"] = relationship("User", back_populates="submissions")
    story: Mapped["Story"] = relationship("Story", back_populates="submissions")

    def __repr__(self):
        """Returns a string representation of the submission.

        Returns
        -------
        str
            A string representation of the submission
        """

        return f'<Submission {self.title!r}>'

    def __str__(self):
        """Returns a string representation of the submission.

        Returns
        -------
        str
            A string representation of the submission
        """

        return f'{self.title}'

    def serialize(self) -> dict:
        """Returns a dictionary representation of the submission.

        Returns
        -------
        dict
            A dictionary representation of the submission
        """

        return {
            'id': self.id,
            'user_id': self.user_id,
            'story_id': self.story_id,
            'submitted_to': self.submitted_to,
            'date_sent': str(self.date_sent),
            'date_reply_received': str(self.date_reply_received),
            'date_published': str(self.date_published),
            'date_paid': str(self.date_paid),
            'result': self.result,
            'amount': self.amount,
            'created': str(self.created),
            'modified': str(self.modified),
        }

    def unserialize(self, data: dict) -> "Submission":
        """Updates the submission's attributes with the values from the dictionary.

        Parameters
        ----------
        data: dict
            The dictionary with the new values for the submission

        Returns
        -------
        Submission
            The unserialized submission
        """

        self.user_id = data.get('user_id', self.user_id)
        self.story_id = data.get('story_id', self.story_id)
        self.submitted_to = data.get('submitted_to', self.submitted_to)
        self.date_sent = data.get('date_sent', self.date_sent)
        self.date_reply_received = data.get('date_reply_received', self.date_reply_received)
        self.date_published = data.get('date_published', self.date_published)
        self.date_paid = data.get('date_paid', self.date_paid)
        self.result = data.get('result', self.result)
        self.amount = data.get('amount', self.amount)
        self.created = data.get('created', self.created)
        self.modified = data.get('modified', self.modified)

        return self

    @validates("submitted_to")
    def validate_submitted_to(self, key, submitted_to: str) -> str:
        """Validates the submitted_to's length.

        Parameters
        ----------
        submitted_to: str
            The submission's recipient

        Returns
        -------
        str
            The validated submitted_to
        """

        if not submitted_to:
            raise ValueError("A submission recipient is required.")

        if len(submitted_to) > 65535:
            raise ValueError("The submission recipient can have no more than 65,535 characters.")

        return submitted_to



import os
from configparser import ConfigParser
from datetime import datetime
from typing import Type, List
from sqlalchemy import or_, func
from sqlalchemy.orm import Session
from gnatwriter.controllers.BaseController import BaseController
from gnatwriter.models import User, Scene, Link, LinkScene, Note, NoteScene, Activity


class SceneController(BaseController):
    """Scene controller encapsulates scene management functionality

    Attributes
    ----------
    _instance : SceneController
        The instance of the scene controller
    _config: ConfigParser
        The configuration parser
    _owner : User
        The current user of the scene controller
    _session : Session
        The database session

    Methods
    -------
    create_scene(story_id: int, chapter_id: int, title: str, description: str, content: str)
        Create a new scene
    update_scene(scene_id: int, title: str, description: str, content: str)
        Update a scene
    change_scene_position(scene_id: int, position: int)
        Change the position of a scene within a chapter
    delete_scene(scene_id: int)
        Delete a scene
    count_scenes_by_chapter_id(chapter_id: int)
        Count the number of scenes associated with a chapter
    get_scene_by_id(scene_id: int)
        Get a scene by id
    get_all_scenes()
        Get all scenes associated with an owner
    get_all_scenes_page(page: int, per_page: int)
        Get a single page of scenes associated with an owner from the database
    get_scenes_by_story_id(story_id: int)
        Get all scenes associated with a story
    get_scenes_page_by_story_id(story_id: int, page: int, per_page: int)
        Get a single page of scenes associated with a story from the database
    get_scenes_by_chapter_id(chapter_id: int)
        Get all scenes associated with a chapter
    get_scenes_page_by_chapter_id(chapter_id: int, page: int, per_page: int)
        Get a single page of scenes associated with a chapter from the database
    search_scenes(search: str)
        Search for scenes by title, description, and content
    append_links_to_scene(scene_id: int, link_ids: list)
        Append links to a scene
    get_links_by_scene_id(scene_id: int)
        Get all links associated with a scene
    append_notes_to_scene(scene_id: int, note_ids: list)
        Append notes to a scene
    get_notes_by_scene_id(scene_id: int)
        Get all notes associated with a scene
    """

    def __init__(
        self, config: ConfigParser, session: Session, owner: Type[User]
    ):
        """Initialize the class"""

        super().__init__(config, session, owner)

    def create_scene(
        self, story_id: int, chapter_id: int, title: str,
        description: str = None, content: str = None
    ) -> Scene:
        """Create a new scene

        Parameters
        ----------
        story_id : int
            The id of the story
        chapter_id : int
            The id of the chapter
        title : str
            The title of the scene
        description : str
            The description of the scene
        content : str
            The content of the scene

        Returns
        -------
        Scene
            The new scene object
        """

        with self._session as session:
            try:
                title_exists = session.query(Scene).filter(
                    Scene.title == title,
                    Scene.chapter_id == chapter_id,
                    Scene.user_id == self._owner.id
                ).first()

                if title_exists:
                    raise Exception('This chapter already has a scene with the same title.')

                position = session.query(
                    func.max(Scene.position)
                ).filter(Scene.chapter_id == chapter_id).scalar()
                position = int(position) + 1 if position else 1
                created = datetime.now()
                modified = created

                scene = Scene(
                    user_id=self._owner.id, story_id=story_id,
                    chapter_id=chapter_id, position=position, title=title,
                    description=description, content=content, created=created,
                    modified=modified
                )

                activity = Activity(
                    user_id=self._owner.id, summary=f'Scene {scene.title[:50]} \
                    created by {self._owner.username}', created=datetime.now()
                )

                session.add(scene)
                session.add(activity)

            except Exception as e:
                session.rollback()
                raise e

            else:
                session.commit()
                return scene

    def update_scene(
        self, scene_id: int, title: str, description: str = None,
        content: str = None
    ) -> Type[Scene]:
        """Update a scene

        Parameters
        ----------
        scene_id : int
            The id of the scene
        title : str
            The title of the scene
        description : str
            The description of the scene
        content : str
            The content of the scene

        Returns
        -------
        bool
            True on success
        """

        with self._session as session:
            try:
                scene = session.query(Scene).filter(
                    Scene.id == scene_id,
                    Scene.user_id == self._owner.id
                ).first()

                if not scene:
                    raise ValueError('Scene not found.')

                scene.title = title
                scene.description = description
                scene.content = content
                scene.modified = datetime.now()

                activity = Activity(
                    user_id=self._owner.id, summary=f'Scene {scene.id} updated \
                    by {self._owner.username}', created=datetime.now()
                )

                session.add(activity)

            except Exception as e:
                session.rollback()
                raise e

            else:
                session.commit()
                return scene

    def change_scene_position(self, scene_id: int, position: int) -> int:
        """Change the position of a scene within a chapter

        First, determine whether the new position is closer to 1 or further from 1. If closer to one, get all sibling
        scenes with positions greater than or equal to the new position but less than the current position, and
        increment those position values by 1. If the target position is further away from 1 than the current position,
        get all sibling scenes with positions greater than the current position but less than or equal to the new
        position, and decrement those position values by 1. Finally, set the position of the target scene to the new
        position. Return the new position.

        Parameters
        ----------
        scene_id : int
            The id of the scene
        position : int
            The position of the scene

        Returns
        -------
        int
            The new position value
        """

        config = ConfigParser()
        project_root = os.path.dirname(os.path.dirname(os.path.dirname(__file__)))
        config.read(f"{project_root}/config.cfg")

        with self._session as session:
            try:
                scene = session.query(Scene).filter(
                    Scene.id == scene_id,
                    Scene.user_id == self._owner.id
                ).first()

                if not scene:
                    raise ValueError('Scene not found.')

                if scene.position == position:
                    return position

                if scene.position > position:
                    scenes = session.query(Scene).filter(
                        Scene.chapter_id == scene.chapter_id,
                        Scene.position >= position,
                        Scene.position < scene.position,
                        Scene.user_id == self._owner.id
                    ).all()
                    for sibling in scenes:
                        sibling.position += 1
                        sibling.created = datetime.strptime(
                            str(sibling.created),
                            config.get("formats", "datetime")
                        )
                        sibling.modified = datetime.now()
                else:
                    scenes = session.query(Scene).filter(
                        Scene.chapter_id == scene.chapter_id,
                        Scene.position > scene.position,
                        Scene.position <= position,
                        Scene.user_id == self._owner.id
                    ).all()
                    for sibling in scenes:
                        sibling.position -= 1
                        sibling.created = datetime.strptime(
                            str(sibling.created),
                            config.get("formats", "datetime")
                        )
                        sibling.modified = datetime.now()

                scene.position = position
                scene.modified = datetime.now()

                activity = Activity(
                    user_id=self._owner.id, summary=f'Scene {scene.title[:50]} \
                    position changed by {self._owner.username}',
                    created=datetime.now()
                )

                session.add(activity)

            except Exception as e:
                session.rollback()
                raise e

            else:
                session.commit()
                return position

    def delete_scene(self, scene_id: int) -> bool:
        """Delete a scene

        Parameters
        ----------
        scene_id : int
            The id of the scene

        Returns
        -------
        bool
            True on success
        """

        with self._session as session:
            try:
                scene = session.query(Scene).filter(
                    Scene.id == scene_id,
                    Scene.user_id == self._owner.id
                ).first()

                if not scene:
                    raise ValueError('Scene not found.')

                siblings = session.query(Scene).filter(
                    Scene.chapter_id == scene.chapter_id,
                    Scene.user_id == self._owner.id,
                    Scene.position > scene.position
                ).all()

                for sibling in siblings:
                    sibling.position -= 1
                    sibling.created = datetime.strptime(
                        str(sibling.created),
                        self._config.get("formats", "datetime")
                    )
                    sibling.modified = datetime.now()

                activity = Activity(
                    user_id=self._owner.id, summary=f'Scene {scene.id} deleted \
                    by {self._owner.username}', created=datetime.now()
                )

                session.delete(scene)
                session.add(activity)

            except Exception as e:
                session.rollback()
                raise e

            else:
                session.commit()
                return True

    def count_scenes_by_chapter_id(self, chapter_id: int) -> int:
        """Count the number of scenes associated with a chapter

        Parameters
        ----------
        chapter_id : int
            The id of the chapter

        Returns
        -------
        int
            The number of scenes
        """

        with self._session as session:

            return session.query(Scene).filter(
                Scene.chapter_id == chapter_id,
                Scene.user_id == self._owner.id
            ).count()

    def get_scene_by_id(self, scene_id: int) -> Type[Scene] | None:
        """Get a scene by id

        Parameters
        ----------
        scene_id : int
            The id of the scene

        Returns
        -------
        Scene
            The scene object
        """

        with self._session as session:

            return session.query(Scene).filter(
                Scene.id == scene_id,
                Scene.user_id == self._owner.id
            ).first()

    def get_all_scenes(self) -> List[Type[Scene]]:
        """Get all scenes associated with an owner

        Scenes are sorted by story id, chapter id, and position in ascending order.

        Returns
        -------
        list
            A list of scene objects
        """

        with self._session as session:
            return session.query(Scene).filter(
                Scene.user_id == self._owner.id
            ).order_by(
                Scene.story_id, Scene.chapter_id, Scene.position
            ).all()

    def get_all_scenes_page(
        self, page: int, per_page: int
    ) -> List[Type[Scene]]:
        """Get a single page of scenes associated with an owner from the database

        Scenes are sorted by story id, chapter id, and position in ascending order

        Parameters
        ----------
        page : int
            The page number
        per_page : int
            The number of rows per page

        Returns
        -------
        list
            A list of scene objects
        """

        with self._session as session:
            offset = (page - 1) * per_page
            return session.query(Scene).filter(
                Scene.user_id == self._owner.id
            ).order_by(
                Scene.story_id, Scene.chapter_id, Scene.position
            ).offset(offset).limit(per_page).all()

    def get_scenes_by_story_id(self, story_id: int) -> List[Type[Scene]]:
        """Get all scenes associated with a story

        Scenes are sorted by chapter id and position in ascending order

        Parameters
        ----------
        story_id : int
            The id of the story

        Returns
        -------
        list
            A list of scene objects
        """

        with self._session as session:
            return session.query(Scene).filter(
                Scene.story_id == story_id,
                Scene.user_id == self._owner.id
            ).order_by(
                Scene.chapter_id, Scene.position
            ).all()

    def get_scenes_page_by_story_id(
        self, story_id: int, page: int, per_page: int
    ) -> List[Type[Scene]]:
        """Get a single page of scenes associated with a story from the database

        Parameters
        ----------
        story_id : int
            The id of the story
        page : int
            The page number
        per_page : int
            The number of rows per page

        Returns
        -------
        list
            A list of scene objects
        """

        with self._session as session:
            offset = (page - 1) * per_page
            return session.query(Scene).filter(
                Scene.story_id == story_id,
                Scene.user_id == self._owner.id
            ).order_by(
                Scene.chapter_id, Scene.position
            ).offset(offset).limit(per_page).all()

    def get_scenes_by_chapter_id(self, chapter_id: int) -> List[Type[Scene]]:
        """Get all scenes associated with a chapter

        Scenes are sorted by position in ascending order

        Parameters
        ----------
        chapter_id : int
            The id of the chapter

        Returns
        -------
        list
            A list of scene objects
        """

        with self._session as session:
            return session.query(Scene).filter(
                Scene.chapter_id == chapter_id,
                Scene.user_id == self._owner.id
            ).order_by(Scene.position).all()

    def get_scenes_page_by_chapter_id(
        self, chapter_id: int, page: int, per_page: int
    ) -> List[Type[Scene]]:
        """Get a single page of scenes associated with a chapter from the database

        Scenes are sorted by position in ascending order

        Parameters
        ----------
        chapter_id : int
            The id of the chapter
        page : int
            The page number
        per_page : int
            The number of rows per page

        Returns
        -------
        list
            A list of scene objects
        """

        with self._session as session:
            offset = (page - 1) * per_page
            return session.query(Scene).filter(
                Scene.chapter_id == chapter_id,
                Scene.user_id == self._owner.id
            ).order_by(Scene.position).offset(offset).limit(per_page).all()

    def search_scenes(self, search: str) -> List[Type[Scene]]:
        """Search for scenes by title, description, and content

        Parameters
        ----------
        search : str
            The search string

        Returns
        -------
        list
            A list of scene objects
        """

        with self._session as session:
            return session.query(Scene).filter(
                or_(
                    Scene.title.like(f'%{search}%'),
                    Scene.description.like(f'%{search}%'),
                    Scene.content.like(f'%{search}%')
                ),
                Scene.user_id == self._owner.id
            ).all()

    def append_links_to_scene(
        self, scene_id: int, link_ids: list
    ) -> Type[Scene]:
        """Append links to a scene

        Parameters
        ----------
        scene_id : int
            The id of the scene
        link_ids : list
            A list of link ids

        Returns
        -------
        Scene
            The updated scene object
        """

        with self._session as session:
            try:
                scene = session.query(Scene).filter(
                    Scene.id == scene_id,
                    Scene.user_id == self._owner.id
                ).first()

                if not scene:
                    raise ValueError('Scene not found.')

                for link_id in link_ids:
                    link = session.query(Link).filter(
                        Link.id == link_id,
                        Link.user_id == self._owner.id
                    ).first()

                    if not link:
                        raise ValueError('Link not found.')

                    link_scene = LinkScene(
                        user_id=self._owner.id, link_id=link_id,
                        scene_id=scene_id, created=datetime.now()
                    )

                    activity = Activity(
                        user_id=self._owner.id, summary=f'Links appended to \
                        scene {scene.id} by {self._owner.username}',
                        created=datetime.now()
                    )

                    session.add(link_scene)
                    session.add(activity)

            except Exception as e:
                session.rollback()
                raise e

            else:
                session.commit()
                return scene

    def get_links_by_scene_id(self, scene_id: int) -> List[Type[Link]]:
        """Get all links associated with a scene

        Parameters
        ----------
        scene_id : int
            The id of the scene

        Returns
        -------
        list
            A list of link objects
        """

        with self._session as session:
            return session.query(Link).join(LinkScene).filter(
                LinkScene.scene_id == scene_id,
                LinkScene.user_id == self._owner.id
            ).all()

    def append_notes_to_scene(
        self, scene_id: int, note_ids: list
    ) -> Type[Scene]:
        """Append notes to a scene

        Parameters
        ----------
        scene_id : int
            The id of the scene
        note_ids : list
            A list of note ids

        Returns
        -------
        Scene
            The updated scene object
        """

        with self._session as session:
            try:
                scene = session.query(Scene).filter(
                    Scene.id == scene_id,
                    Scene.user_id == self._owner.id
                ).first()

                if not scene:
                    raise ValueError('Scene not found.')

                for note_id in note_ids:
                    note = session.query(Note).filter(
                        Note.id == note_id,
                        Note.user_id == self._owner.id
                    ).first()

                    if not note:
                        raise ValueError('Note not found.')

                    note_scene = NoteScene(
                        user_id=self._owner.id, note_id=note_id,
                        scene_id=scene_id, created=datetime.now()
                    )

                    activity = Activity(
                        user_id=self._owner.id, summary=f'Notes appended to \
                        scene {scene.id} by {self._owner.username}',
                        created=datetime.now()
                    )

                    session.add(note_scene)
                    session.add(activity)

            except Exception as e:
                session.rollback()
                raise e

            else:
                session.commit()
                return scene

    def get_notes_by_scene_id(self, scene_id: int) -> List[Type[Note]]:
        """Get all notes associated with a scene

        Parameters
        ----------
        scene_id : int
            The id of the scene

        Returns
        -------
        list
            A list of note objects
        """

        with self._session as session:
            return session.query(Note).join(NoteScene).filter(
                NoteScene.scene_id == scene_id,
                NoteScene.user_id == self._owner.id
            ).all()



import base64
import os
import shutil
import uuid
from configparser import ConfigParser
from datetime import datetime
from typing import Type, Optional, Union, List, Literal, Mapping, Any
from langchain_community.document_loaders.directory import DirectoryLoader
from langchain_community.document_loaders.text import TextLoader
from langchain_community.embeddings.ollama import OllamaEmbeddings
from langchain_community.vectorstores.chroma import Chroma
from langchain_text_splitters import RecursiveCharacterTextSplitter
from ollama import Client, Message
from sqlalchemy.orm import Session
from gnatwriter.controllers.BaseController import BaseController
from gnatwriter.models import User, Assistance, Activity, OllamaModel


class AssistantController(BaseController):
    """Assistant Controller

    Attributes
    ----------
    _config : ConfigParser
        The configuration parser.
    _session_uuid : str
        The UUID of the session.
    _client : OllamaClient
        The Ollama client to be used when making requests.
    _chat_model : str
        The model to be used when chatting.
    _chat_num_ctx : int
        The number of tokens to use as context for the model.
    _chat_keep_alive : Union[float, str]
        The duration to keep the model in memory.
    _generative_model : str
        The model to be used when generating text.
    _generative_num_ctx : int
        The number of tokens to use as context for the model.
    _generative_keep_alive : Union[float, str]
        The duration to keep the model in memory.
    _multimodal_model : str
        The model to be used when describing images.
    _multimodal_num_ctx : int
        The number of tokens to use as context for the model.
    _multimodal_keep_alive : Union[float, str]
        The duration to keep the model in memory.
    _templates : dict
        The templates to be used when generating text.

    Methods
    -------
    chat(
        prompt: str,
        temperature: Optional[float] = 0.5,
        seed: Optional[int] = None,
        priming: str = None,
        options: Optional[dict] = None,
        session_uuid: str = None,
        keep_alive: Optional[Union[float, str]] = None
    )
    generate(
        prompt: str = None,
        temperature: Optional[float] = 0.5,
        seed: Optional[int] = None,
        priming: Optional[str] = None,
        options: Optional[dict] = None,
        session_uuid: str = None,
        keep_alive: Optional[Union[float, str]] = None
    )
        Generate text based on a prompt.
    describe_image(
        images: List[str],
        prompt: Optional[str] = "Describe the image",
        temperature: Optional[float] = 0.5,
        seed: Optional[int] = None,
        priming: Optional[str] = None,
        options: Optional[dict] = None,
        session_uuid: str = None,
        keep_alive: Optional[Union[float, str]] = None
    )
        Describe the contents of an image using the Ollama API.
    get_by_session_uuid(session_uuid: str)
        Get all Assistance messages by session UUID.
    """

    _templates = {}

    def __init__(
            self, config: ConfigParser, session: Session, owner: Type[User]
    ):
        """Initialize the class"""

        super().__init__(config, session, owner)

        uuid4 = str(uuid.uuid4())
        uuid_exists = session.query(Assistance).filter(
            Assistance.session_uuid == uuid4
        ).first()

        while uuid_exists:
            uuid4 = str(uuid.uuid4())
            uuid_exists = session.query(Assistance).filter(
                Assistance.session_uuid == uuid4
            ).first()

        ollama_url = self._config.read("ollama", "url")

        self._client = Client(host=ollama_url)  # If I make this a string instead of a bytearray, the ollama code breaks
        self._session_uuid = uuid4
        self._chat_model = config.get("ollama", "chat_model")
        self._chat_num_ctx = config.getint("ollama", "chat_context_window")
        self._chat_keep_alive = config.get("ollama", "chat_memory_duration")
        self._generative_model = config.get("ollama", "generative_model")
        self._generative_num_ctx = config.getint("ollama", "generative_context_window")
        self._generative_keep_alive = config.get("ollama", "generative_memory_duration")
        self._multimodal_model = config.get("ollama", "multimodal_model")
        self._multimodal_num_ctx = config.getint("ollama", "multimodal_context_window")
        self._multimodal_keep_alive = config.get("ollama", "multimodal_memory_duration")

        self.update_models()

    def update_models(self) -> bool:
        """Update the database with any new models in the list provided by the Ollama API.
        """

        try:

            olist = self._client.list()

            if olist:

                for model in olist["models"]:
                    model_exists = self._session.query(OllamaModel).filter(
                        OllamaModel.model == model["model"]
                    ).first()

                    if not model_exists:

                        details = self._client.show(model["model"])

                        description = details["modelfile"] if details.get("modelfile") else None
                        parameters = details["parameters"] if details.get("parameters") else None
                        template = details["template"] if details.get("template") else None
                        priming = details["system"] if details.get("system") else None

                        created = datetime.now()
                        modified = created

                        ollama_model = OllamaModel(
                            title=model["model"], model=model["model"],
                            description=description, template=template,
                            example=None, priming=priming, params=parameters,
                            created=created, modified=modified
                        )

                        self._session.add(ollama_model)
                        self._session.commit()

                # delete models from db where existing model does not appear in
                # the olist

                with self._session as session:

                    try:

                        models = session.query(OllamaModel).all()

                        for model in models:
                            model_exists = False

                            for omodel in olist["models"]:
                                if model.model == omodel["model"]:
                                    model_exists = True
                                    break

                            if not model_exists:
                                session.delete(model)

                    except Exception as e:
                        session.rollback()
                        return False

                    else:
                        session.commit()

        except Exception as e:
            self._session.rollback()
            return False

        return True

    def chat(
        self,
        prompt: str,
        temperature: Optional[float] = 0.5,
        seed: Optional[int] = None,
        priming: str = None,
        options: Optional[dict] = None,
        session_uuid: str = None,
        keep_alive: Optional[Union[float, str]] = None
    ):
        """Chat with the Chat Assistant.

        The temperature parameter is a float value between 0 and 1. Anything
        greater than 0 will make the assistant more creative, but also more
        unpredictable. The seed parameter is an integer value that can be used
        to make the assistant's responses more predictable. If the temperature
        is set to 0, then the assistant's response will be reproducible, given
        the same seed value.

        Parameters
        ----------
        prompt : str
            The prompt to be used when chatting with the assistant.
        temperature : Optional[float]
            The temperature to be used when making the request. Defaults to
            0.5.
        seed : Optional[int]
            The seed to be used when making the request. Defaults to None.
        priming : Optional[str]
            The priming to be used when chatting with the assistant. Defaults to
            None.
        options : Optional[dict]
            The options to be used when making the request.
        session_uuid : str
            The UUID of the LM session to be used when making the request.
        keep_alive : Optional[Union[float, str]]
            The keep alive value to be used when making the request.
        """

        with self._session as session:

            model = session.query(OllamaModel).filter(
                OllamaModel.model == self._chat_model
            ).first()

        messages = []

        if model:
            if priming is not None:
                messages.append(Message(role="system", content=priming))

        session_uuid = self._session_uuid if not session_uuid else session_uuid

        with self._session as session:

            assistances = session.query(Assistance).filter_by(
                session_uuid=session_uuid
            ).order_by(Assistance.created).all()

            if assistances:
                for assistance in assistances:
                    messages.append(Message(
                        role="user", content=assistance.prompt
                    ))
                    messages.append(Message(
                        role="assistant", content=assistance.content
                    ))

        messages.append(Message(role="user", content=prompt))

        if not options:
            options = {
                "temperature": temperature,
                "num_ctx": self._chat_num_ctx
            }

        if not options.get("temperature"):
            options["temperature"] = temperature

        if not options.get("num_ctx"):
            options["num_ctx"] = self._chat_num_ctx

        keep_alive = self._chat_keep_alive if not keep_alive else keep_alive

        with self._session as session:

            try:

                response = self._client.chat(
                    model=self._chat_model,
                    messages=messages,
                    format='',
                    options=options,
                    keep_alive=keep_alive
                )

                assistance = Assistance(
                    user_id=self._owner.id,
                    session_uuid=session_uuid,
                    assistant="Chat Assistant",
                    model=self._chat_model,
                    priming=priming,
                    prompt=prompt,
                    temperature=temperature,
                    seed=seed,
                    content=response["message"]["content"] if response.get("message") else None,
                    done=response["done"],
                    total_duration=response["total_duration"] if response.get("total_duration") else None,
                    load_duration=response["load_duration"] if response.get("load_duration") else None,
                    prompt_eval_count=response["prompt_eval_count"] if response.get("prompt_eval_count") else None,
                    prompt_eval_duration=response["prompt_eval_duration"] if response.get("prompt_eval_duration") else None,
                    eval_count=response["eval_count"] if response.get("eval_count") else None,
                    eval_duration=response["eval_duration"] if response.get("eval_duration") else None,
                    created=datetime.now()
                )

                summary = f"{self._owner.username} used the Chat Assistant"
                activity = Activity(
                    user_id=self._owner.id, summary=summary,
                    created=datetime.now()
                )

                session.add(assistance)
                session.add(activity)

            except Exception as e:
                session.rollback()
                raise e

            else:
                session.commit()
                return response

    def rag_chat(
        self,
        prompt: str,
        documents: List[str],
        temperature: Optional[float] = 0.5,
        seed: Optional[int] = None,
        priming: str = None,
        options: Optional[dict] = None,
        session_uuid: str = None,
        keep_alive: Optional[Union[float, str]] = None
    ):
        with self._session as session:

            model = session.query(OllamaModel).filter(
                OllamaModel.model == self._chat_model
            ).first()

        messages = []

        if model:
            if priming is not None:
                messages.append(Message(role="system", content=priming))

        session_uuid = self._session_uuid if not session_uuid else session_uuid

        with self._session as session:

            assistances = session.query(Assistance).filter_by(
                session_uuid=session_uuid
            ).order_by(Assistance.created).all()

            if assistances:
                for assistance in assistances:
                    messages.append(Message(
                        role="user", content=assistance.prompt
                    ))
                    messages.append(Message(
                        role="assistant", content=assistance.content
                    ))

        if len(documents) > 1:

            noveler_root = os.path.dirname(os.path.abspath(__file__))
            tmp_dir = f"{noveler_root}/tmp"
            session_tmp_dir = f"{tmp_dir}/{self._session_uuid}"

            if not os.path.exists(session_tmp_dir):
                os.makedirs(session_tmp_dir)

            for document in documents:
                filename = os.path.basename(document).split('/')[-1]
                if os.path.isfile(document) and not os.path.isfile(f"{session_tmp_dir}/{filename}"):
                    shutil.copy(document, f"{session_tmp_dir}/{filename}")

            loader = DirectoryLoader(path=session_tmp_dir, load_hidden=False)

        else:

            loader = TextLoader(file_path=documents[0])

        input_docs = loader.load()
        text_splitter = RecursiveCharacterTextSplitter(
            chunk_size=1000, chunk_overlap=0
        )
        splits = text_splitter.split_documents(documents=input_docs)

        # Create Ollama embeddings and vector store
        embeddings = OllamaEmbeddings(model=f"{self._chat_model}")
        vectorstore = Chroma.from_documents(
            documents=splits, embedding=embeddings
        )

        # Create the retriever
        retriever = vectorstore.as_retriever()
        retrieved_docs = retriever.invoke(prompt)
        formatted_context = "\n\n".join(doc.page_content for doc in retrieved_docs)
        formatted_prompt = f"Question: {prompt}\n\nContext: {formatted_context}"

        messages.append(Message(role="user", content=formatted_prompt))

        if not options:
            options = {
                "temperature": temperature,
                "num_ctx": self._chat_num_ctx
            }

        if not options.get("temperature"):
            options["temperature"] = temperature

        if not options.get("num_ctx"):
            options["num_ctx"] = self._chat_num_ctx

        keep_alive = self._chat_keep_alive if not keep_alive else keep_alive

        with self._session as session:

            try:

                response = self._client.chat(
                    model=self._chat_model,
                    messages=messages,
                    format='',
                    options=options,
                    keep_alive=keep_alive
                )

                assistance = Assistance(
                    user_id=self._owner.id,
                    session_uuid=session_uuid,
                    assistant="RAG Chat Assistant",
                    model=self._chat_model,
                    priming=priming,
                    prompt=prompt,
                    temperature=temperature,
                    seed=seed,
                    content=response["message"]["content"] if response.get("message") else None,
                    done=response["done"],
                    total_duration=response["total_duration"] if response.get("total_duration") else None,
                    load_duration=response["load_duration"] if response.get("load_duration") else None,
                    prompt_eval_count=response["prompt_eval_count"] if response.get("prompt_eval_count") else None,
                    prompt_eval_duration=response["prompt_eval_duration"] if response.get("prompt_eval_duration") else None,
                    eval_count=response["eval_count"] if response.get("eval_count") else None,
                    eval_duration=response["eval_duration"] if response.get("eval_duration") else None,
                    created=datetime.now()
                )

                summary = f"{self._owner.username} used the RAG Chat Assistant"
                activity = Activity(
                    user_id=self._owner.id, summary=summary,
                    created=datetime.now()
                )

                session.add(assistance)
                session.add(activity)

            except Exception as e:
                session.rollback()
                raise e

            else:
                session.commit()
                return response

    def describe_image(
            self,
            images: List[str],
            prompt: Optional[str] = "Describe the image",
            temperature: Optional[float] = 0.5,
            seed: Optional[int] = None,
            priming: Optional[str] = None,
            return_format: Literal['', 'json'] = '',
            options: Optional[dict] = None,
            session_uuid: str = None,
            keep_alive: Optional[Union[float, str]] = None
    ):
        """Describe the contents of an image using the Ollama API.

        Parameters
        ----------
        images : List[str]
            A list of image file paths to be described.
        temperature : Optional[float]
            The temperature to be used when making the request. Defaults to
            0.5.
        seed : Optional[int]
            The seed to be used when making the request. Defaults to None.
        prompt : Optional[str]
            The prompt to be used when requesting the description. Defaults to
            "Describe the contents of the image:".
        priming : Optional[str]
            The priming to be used when describing the image.
        return_format : str
            The format to be used to return the request.
        options : Optional[dict]
            The options to be used when making the request.
        session_uuid : str
            The UUID of the session to be used when making the request.
        keep_alive : Optional[Union[float, str]]
            The keep alive value to be used when making the request.
        """
        encoded_images = []

        for image in images:
            with open(image, "rb") as file:
                encoded_images.append(base64.b64encode(file.read()).decode())

        template = [template for template in self._templates if template.model == self._multimodal_model][0]

        if not priming:
            priming = template.priming

        session_uuid = self._session_uuid if not session_uuid else session_uuid

        if not options:
            options = {
                "temperature": temperature,
                "num_ctx": self._multimodal_num_ctx
            }

        if not options.get("temperature"):
            options["temperature"] = temperature

        if not options.get("num_ctx"):
            options["num_ctx"] = self._multimodal_num_ctx

        keep_alive = self._multimodal_keep_alive if not keep_alive else keep_alive

        with self._session as session:

            try:

                response = self._client.generate(
                    model=self._multimodal_model, prompt=prompt, system=priming,
                    template=template.template, context=None, stream=False,
                    raw=False, format=return_format, images=encoded_images,
                    options=options, keep_alive=keep_alive
                )

                content = response["response"] if response.get("response") else None

                assistance = Assistance(
                    user_id=self._owner.id,
                    session_uuid=session_uuid,
                    assistant="Multimodal Assistant",
                    model=self._multimodal_model,
                    priming=priming,
                    prompt=prompt,
                    temperature=temperature,
                    seed=seed,
                    content=content,
                    done=response["done"],
                    total_duration=response["total_duration"] if response.get("total_duration") else None,
                    load_duration=response["load_duration"] if response.get("load_duration") else None,
                    prompt_eval_count=response["prompt_eval_count"] if response.get("prompt_eval_count") else None,
                    prompt_eval_duration=response["prompt_eval_duration"] if response.get("prompt_eval_duration") else None,
                    eval_count=response["eval_count"] if response.get("eval_count") else None,
                    eval_duration=response["eval_duration"] if response.get("eval_duration") else None,
                    created=datetime.now()
                )

                summary = f"{self._owner.username} used the Multimodal Assistant"
                activity = Activity(
                    user_id=self._owner.id, summary=summary,
                    created=datetime.now()
                )

                session.add(assistance)
                session.add(activity)

            except Exception as e:
                session.rollback()
                raise e

            else:
                session.commit()
                return response

    def generate(
            self,
            prompt: str = None,
            temperature: Optional[float] = 0.5,
            seed: Optional[int] = None,
            priming: Optional[str] = None,
            return_format: Literal['', 'json'] = '',
            options: Optional[dict] = None,
            session_uuid: str = None,
            keep_alive: Optional[Union[float, str]] = None
    ):
        """Describe the contents of an image using the Ollama API.

        Parameters
        ----------
        prompt : str
            The prompt to be used when requesting the description.
        temperature : Optional[float]
            The temperature to be used when making the request. Defaults to
            0.5.
        seed : Optional[int]
            The seed to be used when making the request. Defaults to None.
        priming : Optional[str]
            The priming to be used prior to the prompt.
        return_format : str
            The format to be used to return the request.
        options : Optional[dict]
            The options to be used when making the request.
        session_uuid : str
            The UUID of the session to be used when making the request.
        keep_alive : Optional[Union[float, str]]
            The keep alive value to be used when making the request.
        """
        template = [template for template in self._templates if template.model == self._generative_model][0]

        if not priming:
            priming = """Generative Assistant is ready to perform a task."""

        session_uuid = self._session_uuid if not session_uuid else session_uuid

        if not options:
            options = {
                "temperature": temperature,
                "num_ctx": self._generative_num_ctx
            }

        if not options.get("temperature"):
            options["temperature"] = temperature

        if not options.get("num_ctx"):
            options["num_ctx"] = self._generative_num_ctx

        keep_alive = self._generative_keep_alive if not keep_alive else keep_alive

        with self._session as session:

            try:

                response = self._client.generate(
                    model=self._generative_model, prompt=prompt, system=priming,
                    template=template.template, context=None, stream=False,
                    raw=False, format=return_format, images=None,
                    options=options, keep_alive=keep_alive
                )

                assistance = Assistance(
                    user_id=self._owner.id,
                    session_uuid=session_uuid,
                    assistant="Generative Assistant",
                    model=self._generative_model,
                    priming=priming,
                    prompt=prompt,
                    temperature=temperature,
                    seed=seed,
                    content=response["response"] if response.get("response") else None,
                    done=response["done"],
                    total_duration=response["total_duration"] if response.get("total_duration") else None,
                    load_duration=response["load_duration"] if response.get("load_duration") else None,
                    prompt_eval_count=response["prompt_eval_count"] if response.get("prompt_eval_count") else None,
                    prompt_eval_duration=response["prompt_eval_duration"] if response.get("prompt_eval_duration") else None,
                    eval_count=response["eval_count"] if response.get("eval_count") else None,
                    eval_duration=response["eval_duration"] if response.get("eval_duration") else None,
                    created=datetime.now()
                )

                summary = f"{self._owner.username} used the Generative Assistant"
                activity = Activity(
                    user_id=self._owner.id, summary=summary,
                    created=datetime.now()
                )

                session.add(assistance)
                session.add(activity)

            except Exception as e:
                session.rollback()
                raise e

            else:
                session.commit()
                return response

    def list_models(self) -> Mapping[str, Any]:
        """List all available models."""
        return self._client.list()

    def show_model(self, model: str) -> Mapping[str, Any]:
        """Show the details of a specific model."""
        return self._client.show(model)

    def __str__(self):
        """Return the class string representation."""
        return f"GnatWriter Application Assistant Controller"

    def __repr__(self):
        """Return the class representation."""
        return f"{self.__class__.__name__}()"



from datetime import datetime
from sqlalchemy import Integer, ForeignKey, DateTime
from sqlalchemy.orm import Mapped, mapped_column, relationship
from gnatwriter.models import User, Link, Location, Base


class LinkLocation(Base):
    """The LinkLocation class represents the relationship between a link and a location.

    Attributes
    ----------
        user_id: int
            The id of the owner of this entry
        link_id: int
            The link's id
        location_id: int
            The location's id
        created: str
            The creation datetime of the link between the Link and the Location
        user: User
            The user who owns this entry
        link: Link
            The link that the location belongs to
        location: Location
            The location that the link belongs to

    Methods
    -------
        __repr__()
            Returns a string representation of the relationship
        __str__()
            Returns a string representation of the relationship
        serialize()
            Returns a dictionary representation of the relationship
        unserialize(data: dict)
            Updates the relationship's attributes with the values from the dictionary
    """

    __tablename__ = 'links_locations'
    user_id: Mapped[int] = mapped_column(Integer, ForeignKey('users.id'))
    link_id: Mapped[int] = mapped_column(Integer, ForeignKey('links.id'), primary_key=True)
    location_id: Mapped[int] = mapped_column(Integer, ForeignKey('locations.id'), primary_key=True)
    created: Mapped[str] = mapped_column(DateTime, default=str(datetime.now()))
    user: Mapped["User"] = relationship("User")
    link: Mapped["Link"] = relationship("Link", back_populates="locations")
    location: Mapped["Location"] = relationship("Location", back_populates="links")

    def __repr__(self):
        """Returns a string representation of the relationship.

        Returns
        -------
        str
            A string representation of the relationship
        """

        return f'<LinkLocation {self.link.title!r} - {self.location.title!r}>'

    def __str__(self):
        """Returns a string representation of the relationship.

        Returns
        -------
        str
            A string representation of the relationship
        """

        return f'{self.link.title} - {self.location.title}'

    def serialize(self) -> dict:
        """Returns a dictionary representation of the relationship.

        Returns
        -------
        dict
            A dictionary representation of the relationship
        """

        return {
            'user_id': self.user_id,
            'link_id': self.link_id,
            'location_id': self.location_id,
            'created': str(self.created),
        }

    def unserialize(self, data: dict) -> "LinkLocation":
        """Updates the relationship's attributes with the values from the dictionary.

        Parameters
        ----------
        data: dict
            The dictionary with the new values for the relationship

        Returns
        -------
        LinkLocation
            The unserialized relationship
        """

        self.user_id = data.get('user_id', self.user_id)
        self.link_id = data.get('link_id', self.link_id)
        self.location_id = data.get('location_id', self.location_id)
        self.created = data.get('created', self.created)

        return self



import json
import os
from configparser import ConfigParser
from datetime import datetime
from typing import Type
from sqlalchemy.orm import Session
from gnatwriter.controllers import BaseController
from gnatwriter.models import User, Story, Activity, Character, Event


class ExportController(BaseController):
    """Export controller encapsulates export functionality"""

    _export_root: str

    def __init__(
            self, config: ConfigParser, session: Session, owner: Type[User]
    ):
        """Initialize the class"""

        super().__init__(config, session, owner)

        try:

            export_root = self._config.get("export", "root")
            self._export_root = export_root

        except Exception as e:
            raise e

    def export_story_to_json(self, story_id: int) -> bool:
        """Export a story to a JSON file

        This method exports a story to a JSON file and stores that file in a
        folder named with the story id. So, for example, if the story id is 1,
        the file would be stored in the folder export/<user UUID>/1/story.json.

        Parameters
        ----------
        story_id : int
            The id of the story to export

        Returns
        -------
        bool
            True if successful, False otherwise
        """

        user_folder = f"{self._export_root}/{self._owner.uuid}"

        if not os.path.exists(user_folder):
            os.makedirs(user_folder)

        with self._session as session:

            story = session.query(Story).filter(
                Story.id == story_id,
                Story.user_id == self._owner.id
            ).first()

            if not story:
                return False

            json_story = json.dumps(story.serialize(), indent=4)
            story_folder = f"{user_folder}/stories/{story_id}"

            if not os.path.exists(story_folder):
                os.makedirs(story_folder)

            story_file = f"{story_folder}/story.json"

            with open(story_file, "w") as output:

                output.write(json_story)

        with self._session as session:

            try:

                activity = Activity(
                    user_id=self._owner.id, summary=f'Story exported to JSON by \
                    {self._owner.username}', created=datetime.now()
                )

                session.add(activity)

            except Exception as e:
                session.rollback()
                raise e

            else:
                session.commit()

        return True

    def export_story_to_text(self, story_id: int) -> bool:
        """Export a story to a text file

        This method exports a story to a text file and stores that file in a
        folder named with the story id. So, for example, if the story id is 1,
        the file would be stored in the folder export/<user UUID>/1/story.txt.

        Parameters
        ----------
        story_id : int
            The id of the story to export

        Returns
        -------
        bool
            True if successful, False otherwise
        """

        user_folder = f"{self._export_root}/{self._owner.uuid}"

        if not os.path.exists(user_folder):
            os.makedirs(user_folder)

        with self._session as session:

            story = session.query(Story).filter(
                Story.id == story_id,
                Story.user_id == self._owner.id
            ).first()

            if not story:
                return False

            story_folder = f"{user_folder}/stories/{story_id}"

            if not os.path.exists(story_folder):
                os.makedirs(story_folder)

            story_file = f"{story_folder}/story.txt"
            dict_story = story.serialize()

            with open(story_file, "w") as output:

                output.write(f"{dict_story['title']}\n")
                output.write(f"{dict_story['description']}\n")

                author_count = len(dict_story["authors"])

                counter = 0
                is_first = True
                author_string = ""

                for author in dict_story["authors"]:

                    counter += 1

                    if is_first:
                        author_string += f"By {author['name']}"
                        is_first = False

                    elif counter == (author_count - 2):
                        author_string += f", {author.name}"

                    elif counter == (author_count - 1):
                        author_string += f" and {author.name}"

                output.write(f"{author_string}\n")

                for chapter in dict_story["chapters"]:

                    output.write(f"\n\n{chapter['title']}\n\n")

                    if chapter['description']:
                        output.write(f"\n\n{chapter['description']}\n\n")

                    for scene in chapter["scenes"]:

                        if chapter['title']:
                            output.write(f"{scene['title']}\n")

                        if scene['content']:
                            output.write(f"{scene['content']}\n")

        with self._session as session:

            try:

                activity = Activity(
                    user_id=self._owner.id, summary=f'Story exported to TEXT by \
                    {self._owner.username}', created=datetime.now()
                )

                session.add(activity)

            except Exception as e:
                session.rollback()
                raise e

            else:
                session.commit()

        return True

    def export_character_to_json(self, character_id: int) -> bool:
        """Export a character to a JSON file

        Parameters
        ----------
        character_id : int
            The id of the character to export

        Returns
        -------
        bool
            True if successful, False otherwise
        """

        user_folder = f"{self._export_root}/{self._owner.uuid}"

        if not os.path.exists(user_folder):
            os.makedirs(user_folder)

        with self._session as session:

            character = session.query(Character).filter(
                Character.id == character_id,
                Character.user_id == self._owner.id
            ).first()

            if not character:
                return False

            json_character = json.dumps(character.serialize(), indent=4)
            character_folder = f"{user_folder}/characters/{character_id}"

            if not os.path.exists(character_folder):
                os.makedirs(character_folder)

            character_file = f"{character_folder}/character.json"

            with open(character_file, "w") as output:

                output.write(json_character)

        with self._session as session:

            try:

                activity = Activity(
                    user_id=self._owner.id, summary=f'Character exported to \
                    JSON by {self._owner.username}', created=datetime.now()
                )

                session.add(activity)

            except Exception as e:
                session.rollback()
                raise e

            else:
                session.commit()

        return True

    def export_character_to_text(self, character_id: int) -> bool:
        """Export a character to a text file

        Parameters
        ----------
        character_id : int
            The id of the character to export

        Returns
        -------
        bool
            True if successful, False otherwise
        """

        user_folder = f"{self._export_root}/{self._owner.uuid}"

        if not os.path.exists(user_folder):
            os.makedirs(user_folder)

        with self._session as session:

            character = session.query(Character).filter(
                Character.id == character_id,
                Character.user_id == self._owner.id
            ).first()

            if not character:
                return False

            character_folder = f"{user_folder}/characters/{character_id}"

            if not os.path.exists(character_folder):
                os.makedirs(character_folder)

            character_file = f"{character_folder}/character.txt"
            dict_character = character.serialize()

            with open(character_file, "w") as output:

                output.write(
                    f"Character Profile for {dict_character['full_name']}"
                )

                if dict_character['description']:
                    output.write(f"\n\n{dict_character['description']}")

                if dict_character['wounds']:
                    output.write(f"\n\nThe character's wounds:\n{dict_character['wounds']}")

                if dict_character["gender"]:
                    output.write(f"\n\nGender: {dict_character['gender']}")

                if dict_character["sex"]:
                    output.write(f"\nSex: {dict_character['sex']}")

                if dict_character["ethnicity"]:
                    output.write(f"\nEthnicity: {dict_character['ethnicity']}")

                if dict_character["race"]:
                    output.write(f"\nRace: {dict_character['race']}")

                if dict_character["tribe_or_clan"]:
                    output.write(f"\nTribe or Clan: {dict_character['tribe_or_clan']}")

                if dict_character["nationality"]:
                    output.write(f"\nNationality: {dict_character['nationality']}")

                if dict_character["religion"]:
                    output.write(f"\nReligion: {dict_character['religion']}")

                if dict_character["mbti"]:
                    output.write(f"\nMyers-Briggs Type Indicator: {dict_character['mbti']}")

                if dict_character["enneagram"]:
                    output.write(f"\nEnneagram Type: {dict_character['enneagram']}")

                if dict_character["occupation"]:
                    output.write(f"\nOccupation: {dict_character['occupation']}")

                if dict_character["education"]:
                    output.write(f"\nEducation: {dict_character['education']}")

                if dict_character["marital_status"]:
                    output.write(f"\nMarital Status: {dict_character['marital_status']}")

                if dict_character["children"]:
                    output.write(f"\nThis character has or had children.")
                else:
                    output.write(f"\nThis character has no children.")

                if dict_character["age"]:
                    output.write(f"\nAge: {dict_character['age']}")

                if dict_character["date_of_birth"]:
                    output.write(f"\nDate of Birth: {dict_character['date_of_birth']}")

                if dict_character["date_of_death"]:
                    output.write(f"\nDate of Death: {dict_character['date_of_death']}")

                if len(dict_character["traits"]) > 0:
                    output.write("\n\nTraits and Magnitudes on 0-100 Scale (0 = undetectable, 100 = maximum level):")
                    for trait in dict_character["traits"]:
                        output.write(f"\n{trait['name']} - {trait['magnitude']}")

                if dict_character["relationships"]:
                    output.write("\n\nRelationships to other Characters:")
                    for relationship in dict_character["relationships"]:
                        output.write(
                            f"""\n\nRelationship to {relationship['related_name']} in the category of \
{relationship['relationship_type']}:
Start Date: {relationship['start_date']}
End Date: {relationship['end_date']}
                            """
                        )
                        output.write(
                            f"\nDescription:  {relationship['description']}"
                        )

                if dict_character["events"]:
                    output.write(
                        "\n\nEvents with which this character is associated:"
                    )
                    for character_event in dict_character["events"]:
                        output.write(
                            f"\n{character_event['event']}"
                        )

                if dict_character["stories"]:
                    output.write(
                        "\n\nStories in which this character appears:"
                    )
                    for character_story in dict_character["stories"]:
                        output.write(
                            f"\n{character_story['story_name']}"
                        )

        with self._session as session:

            try:

                activity = Activity(
                    user_id=self._owner.id, summary=f'Character exported to \
                    TEXT by {self._owner.username}', created=datetime.now()
                )

                session.add(activity)

            except Exception as e:
                session.rollback()
                raise e

            else:
                session.commit()

        return True

    def export_event_to_json(self, event_id: int) -> bool:
        """Export an event to a JSON file

        Parameters
        ----------
        event_id : int
            The id of the event to export

        Returns
        -------
        bool
            True if successful, False otherwise
        """

        user_folder = f"{self._export_root}/{self._owner.uuid}"

        if not os.path.exists(user_folder):
            os.makedirs(user_folder)

        with self._session as session:

            event = session.query(Event).filter(
                Event.id == event_id,
                Event.user_id == self._owner.id
            ).first()

            if not event:
                return False

            json_event = json.dumps(event.serialize(), indent=4)
            event_folder = f"{user_folder}/events/{event_id}"

            if not os.path.exists(event_folder):
                os.makedirs(event_folder)

            event_file = f"{event_folder}/event.json"

            with open(event_file, "w") as output:

                output.write(json_event)

        with self._session as session:

            try:

                activity = Activity(
                    user_id=self._owner.id, summary=f'Event exported to JSON by \
                    {self._owner.username}', created=datetime.now()
                )

                session.add(activity)

            except Exception as e:
                session.rollback()
                raise e

            else:
                session.commit()

        return True

    def export_event_to_text(self, event_id: int) -> bool:
        """Export an event to a text file

        Parameters
        ----------
        event_id : int
            The id of the event to export

        Returns
        -------
        bool
            True if successful, False otherwise
        """

        user_folder = f"{self._export_root}/{self._owner.uuid}"

        if not os.path.exists(user_folder):
            os.makedirs(user_folder)

        with self._session as session:

            event = session.query(Event).filter(
                Event.id == event_id,
                Event.user_id == self._owner.id
            ).first()

            if not event:
                return False

            event_folder = f"{user_folder}/events/{event_id}"

            if not os.path.exists(event_folder):
                os.makedirs(event_folder)

            event_file = f"{event_folder}/event.txt"
            dict_event = event.serialize()

            with open(event_file, "w") as output:

                output.write(f"Event report for {dict_event['title']}")

                if dict_event["start_datetime"]:
                    output.write(
                        f"\n\nStart Datetime: {dict_event['start_datetime']}"
                    )

                if dict_event["end_datetime"]:
                    output.write(
                        f"\nEnd Datetime: {dict_event['end_datetime']}"
                    )

                if dict_event["description"]:
                    output.write(f"\n\n{dict_event['description']}")

        with self._session as session:

            try:

                activity = Activity(
                    user_id=self._owner.id, summary=f'Event exported to TEXT by \
                    {self._owner.username}', created=datetime.now()
                )

                session.add(activity)

            except Exception as e:
                session.rollback()
                raise e

            else:
                session.commit()

        return True



from configparser import ConfigParser
from typing import Type
from sqlalchemy.orm import Session
from gnatwriter.models import User


class BaseController:
    """Base controller encapsulates common functionality for all controllers

    Attributes
    ----------
    _instance : BaseController
        The instance of the base controller
    _config : ConfigParser
        The configuration parser
    _owner : User
        The current user of the base controller
    _session : Session
        The database session
    """
    _instance = None
    _config = None
    _owner = None
    _session = None

    def __new__(cls, config: ConfigParser, session: Session, owner: Type[User]):
        """Enforce Singleton pattern"""

        if cls._instance is None:
            cls._instance = super().__new__(cls)

        return cls._instance

    def __init__(
        self, config: ConfigParser, session: Session, owner: Type[User]
    ):
        """Initialize the class"""

        self._config = config
        self._session = session
        self._owner = owner



from datetime import datetime
from sqlalchemy import Integer, ForeignKey, String, DateTime
from sqlalchemy.orm import Mapped, mapped_column, relationship, validates
from gnatwriter.models import User, Character, Base


class CharacterTrait(Base):
    """The CharacterTrait class represents a trait of a character.

    Attributes
    ----------
        id: int
            The trait's id
        user_id: int
            The id of the owner of this entry
        character_id: int
            The character's id
        position: int
            The position of the trait in the character's traits
        name: str
            The name of the trait
        magnitude: int
            The magnitude of the trait expressed as an integer between 0 and 100
        created: str
            The trait's creation date in datetime form: yyy-mm-dd hh:mm:ss
        modified: str
            The trait's last modification date in datetime form: yyy-mm-dd hh:mm:ss

    Methods
    -------
        __repr__()
            Returns a string representation of the trait
        __str__()
            Returns a string representation of the trait
        serialize()
            Returns a dictionary representation of the trait
        unserialize(data: dict)
            Updates the trait's attributes with the values from the dictionary
        validate_name(name: str)
            Validates the name's length
        validate_magnitude(magnitude: int)
            Validates the magnitude's value
    """

    __tablename__ = 'characters_traits'
    id: Mapped[int] = mapped_column(Integer, primary_key=True)
    user_id: Mapped[int] = mapped_column(Integer, ForeignKey('users.id'))
    character_id: Mapped[int] = mapped_column(
        Integer, ForeignKey('characters.id')
    )
    position: Mapped[int] = mapped_column(Integer, nullable=False, default=1)
    name: Mapped[str] = mapped_column(String(100), nullable=False)
    magnitude: Mapped[int] = mapped_column(Integer, nullable=False, default=0)
    created: Mapped[str] = mapped_column(DateTime, default=str(datetime.now()))
    modified: Mapped[str] = mapped_column(
        DateTime, default=str(datetime.now()), onupdate=str(datetime.now())
    )
    user: Mapped["User"] = relationship("User")
    character: Mapped["Character"] = relationship(
        "Character", back_populates="traits", lazy="joined"
    )

    def __repr__(self):
        """Returns a string representation of the trait.

        Returns
        -------
        str
            A string representation of the trait
        """

        return f'{self.character.first_name!r} {self.character.last_name!r} - {self.name!r}'

    def __str__(self):
        """Returns a string representation of the trait.

        Returns
        -------
        str
            A string representation of the trait
        """

        return f'{self.character.first_name} {self.character.last_name} - {self.name}'

    def serialize(self) -> dict:
        """Returns a dictionary representation of the trait.

        Returns
        -------
        dict
            A dictionary representation of the trait
        """

        return {
            'user_id': self.user_id,
            'character_id': self.character_id,
            'position': self.position,
            'name': self.name,
            'magnitude': self.magnitude,
            'created': str(self.created),
            'modified': str(self.modified),
        }

    def unserialize(self, data: dict) -> "CharacterTrait":
        """Updates the trait's attributes with the values from the dictionary.

        Parameters
        ----------
        data: dict
            The dictionary with the new values for the trait

        Returns
        -------
        CharacterTrait
            The unserialized trait
        """

        self.user_id = data.get('user_id', self.user_id)
        self.character_id = data.get('character_id', self.character_id)
        self.position = data.get('position', self.position)
        self.name = data.get('name', self.name)
        self.magnitude = data.get('magnitude', self.magnitude)
        self.created = data.get('created', self.created)
        self.modified = data.get('modified', self.modified)

        return self

    @validates("name")
    def validate_name(self, key, name: str) -> str:
        """Validates the name's length.

        Parameters
        ----------
        name: str
            The trait's name

        Returns
        -------
        str
            The validated name
        """

        if not name:
            raise ValueError("The trait name cannot be empty.")

        return name

    @validates("magnitude")
    def validate_magnitude(self, key, magnitude: int) -> int:
        """Validates the magnitude's value.

        Parameters
        ----------
        magnitude: int
            The trait's magnitude

        Returns
        -------
        str
            The validated magnitude
        """

        if magnitude < 0:
            raise ValueError("The trait magnitude cannot be negative.")

        if magnitude > 100:
            raise ValueError("The trait magnitude cannot be greater than 100.")

        return magnitude



from configparser import ConfigParser
from datetime import datetime
from typing import Type, List
from sqlalchemy.orm import Session
from gnatwriter.controllers import BaseController
from gnatwriter.models import User, OllamaModel, Activity


class OllamaModelController(BaseController):
    """The OllamaModelController class encapsulates the logic for managing
    Ollama models in the database.
    """

    def __init__(
        self, config: ConfigParser, session: Session, owner: Type[User]
    ):
        """Initialize the class"""

        super().__init__(config, session, owner)

    def create_model(
        self, title: str, model: str, description: str = None,
        template: str = None, example: str = None, priming: str = None,
        params: str = None
    ) -> OllamaModel:
        """Create a model

        Parameters
        ----------
        title : str
            The title of the model
        model : str
            The model name
        description : str
            The model description
        template : str
            The model template
        example : str
            The model example
        priming : str
            The model priming
        params : str
            The model params

        Returns
        -------
        OllamaModel
            The model object
        """

        with self._session as session:

            try:

                created = datetime.now()
                modified = created

                model = OllamaModel(
                    title=title, model=model, description=description,
                    template=template, example=example, priming=priming,
                    params=params, created=created, modified=modified
                )

                activity = Activity(
                    user_id=self._owner.id, summary=f'Ollama model \
                    {model.title[:50]} created by {self._owner.username}',
                    created=datetime.now()
                )

                session.add(model)
                session.add(activity)

            except Exception as e:
                session.rollback()
                raise e

            else:
                session.commit()
                return model

    def get_model(self, model: str) -> Type[OllamaModel] | None:
        """Get an activity by id

        Parameters
        ----------

        Returns
        -------
        """

        with self._session as session:

            model = session.query(OllamaModel).filter(
                OllamaModel.model == model
            ).first()

            return model if model else None

    def get_models(self) -> List[Type[OllamaModel]]:
        """Get all models stored in the database

        Returns
        -------
        list
            A list of model objects
        """

        with self._session as session:
            return session.query(OllamaModel).all()




from enum import Enum


class SubmissionResultType(Enum):
    """The SubmissionResultType class represents the possible results or status of a story submission.

    Attributes
    ----------
        ACCEPTED: str
            The submission was accepted
        REJECTED: str
            The submission was rejected
        REWRITE_REQUESTED: str
            The submission was requested to be rewritten
        PENDING: str
            The submission is pending
        WITHDRAWN: str
            The submission was withdrawn
        IGNORED: str
            The submission was ignored
    """

    ACCEPTED = 'Accepted'
    REJECTED = 'Rejected'
    REWRITE_REQUESTED = 'Rewrite Requested'
    PENDING = 'Pending'
    WITHDRAWN = 'Withdrawn'
    IGNORED = 'Ignored'



from configparser import ConfigParser
from datetime import datetime
from typing import Type, List

from gnatwriter.models.NoteStory import NoteStory

from gnatwriter.models.Story import Story
from sqlalchemy import or_
from sqlalchemy.orm import Session
from gnatwriter.controllers.BaseController import BaseController
from gnatwriter.models import User, Note, Activity


class NoteController(BaseController):
    """Note controller encapsulates note management functionality

    Attributes
    ----------
    _instance : NoteController
        The instance of the note controller
    _config: ConfigParser
        The configuration parser
    _owner : User
        The current user of the note controller
    _session : Session
        The database session

    Methods
    -------
    create_note(title: str, content: str)
        Create a new note
    update_note(note_id: int, title: str, content: str)
        Update a note
    delete_note_by_id(note_id: int)
        Delete a note
    get_all_notes()
        Get all notes associated with an owner
    get_all_notes_page(page: int, per_page: int)
        Get a single page of notes associated with an owner from the database
    search_notes(search: str)
        Search for notes by title and content
    get_notes_by_story_id(story_id: int)
        Get all notes associated with a story
    get_notes_page_by_story_id(story_id: int, page: int, per_page: int)
        Get a single page of notes associated with a story from the database
    """

    def __init__(
        self, config: ConfigParser, session: Session, owner: Type[User]
    ):
        """Initialize the class"""

        super().__init__(config, session, owner)

    def create_note(self, title: str, content: str) -> Note:
        """Create a new note

        Parameters
        ----------
        title : str
            The title of the note
        content : str
            The content of the note

        Returns
        -------
        Note
            The new note object
        """

        with self._session as session:

            try:

                created = datetime.now()
                modified = created

                note = Note(
                    user_id=self._owner.id, title=title, content=content,
                    created=created, modified=modified
                )

                activity = Activity(
                    user_id=self._owner.id, summary=f'Note {note.title[:50]} \
                    created by {self._owner.username}', created=datetime.now()
                )

                session.add(note)
                session.add(activity)

            except Exception as e:
                session.rollback()
                raise e

            else:
                session.commit()
                return note

    def update_note(self, note_id: int, title: str, content: str) -> Type[Note]:
        """Update a note

        Parameters
        ----------
        note_id : int
            The id of the note
        title : str
            The title of the note
        content : str
            The content of the note

        Returns
        -------
        Note
            The updated note object
        """

        with self._session as session:
            try:
                note = session.query(Note).filter(
                    Note.id == note_id, Note.user_id == self._owner.id
                ).first()

                if not note:
                    raise ValueError('Note not found.')

                note.title = title
                note.content = content
                note.modified = datetime.now()

                activity = Activity(
                    user_id=self._owner.id, summary=f'Note {note.id} updated \
                    by {self._owner.username}', created=datetime.now()
                )

                session.add(activity)

            except Exception as e:
                session.rollback()
                raise e

            else:
                session.commit()
                return note

    def delete_note_by_id(self, note_id: int) -> bool:
        """Delete a note

        Parameters
        ----------
        note_id : int
            The id of the note

        Returns
        -------
        bool
            True on success
        """

        with self._session as session:
            try:
                note = session.query(Note).filter(
                    Note.id == note_id,
                    Note.user_id == self._owner.id
                ).first()

                if not note:
                    raise ValueError('Note not found.')

                activity = Activity(
                    user_id=self._owner.id, summary=f'Note {note.id} deleted \
                    by {self._owner.username}', created=datetime.now()
                )

                session.delete(note)
                session.add(activity)

            except Exception as e:
                session.rollback()
                raise e

            else:
                session.commit()
                return True

    def get_note_by_id(self, note_id: int) -> Type[Note] | None:
        """Get a note by id

        Parameters
        ----------
        note_id : int
            The id of the note

        Returns
        -------
        Note
            The note object
        """

        with self._session as session:
            return session.query(Note).filter(
                Note.id == note_id,
                Note.user_id == self._owner.id
            ).first()

    def get_all_notes(self) -> List[Type[Note]]:
        """Get all notes associated with an owner

        Returns
        -------
        list
            A list of note objects
        """

        with self._session as session:
            return session.query(Note).filter(
                Note.user_id == self._owner.id
            ).all()

    def get_all_notes_page(self, page: int, per_page: int) -> List[Type[Note]]:
        """Get a single page of notes associated with an owner from the database

        Parameters
        ----------
        page : int
            The page number
        per_page : int
            The number of rows per page

        Returns
        -------
        list
            A list of note objects
        """

        with self._session as session:
            offset = (page - 1) * per_page
            return session.query(Note).filter(
                Note.owner_id == self._owner.id
            ).offset(offset).limit(per_page).all()

    def search_notes(self, search: str) -> List[Type[Note]]:
        """Search for notes by title and content

        Parameters
        ----------
        search : str
            The search string

        Returns
        -------
        list
            A list of note objects
        """

        with self._session as session:
            return session.query(Note).filter(
                or_(
                    Note.title.like(f'%{search}%'),
                    Note.content.like(f'%{search}%')
                ),
                Note.user_id == self._owner.id
            ).all()

    def get_notes_by_story_id(self, story_id: int) -> List[Type[Note]]:
        """Get all notes associated with a story

        Parameters
        ----------
        story_id : int
            The id of the story

        Returns
        -------
        list
            A list of note objects
        """

        with self._session as session:

            return session.query(Note).join(NoteStory).filter(
                NoteStory.story_id == story_id,
                NoteStory.user_id == self._owner.id
            ).all()

    def get_notes_page_by_story_id(self, story_id: int, page: int, per_page: int) -> List[Type[Note]]:
        """Get a single page of notes associated with a story from the database

        Parameters
        ----------
        story_id : int
            The id of the story
        page : int
            The page number
        per_page : int
            The number of rows per page

        Returns
        -------
        list
            A list of note objects
        """

        with self._session as session:

            offset = (page - 1) * per_page

            return session.query(Note).join(NoteStory).filter(
                NoteStory.story_id == story_id,
                NoteStory.user_id == self._owner.id
            ).offset(offset).limit(per_page).all()



import uuid
from configparser import ConfigParser
from datetime import datetime
from typing import Type, List
import bcrypt
from sqlalchemy import func
from sqlalchemy.orm import Session
from gnatwriter.controllers.BaseController import BaseController
from gnatwriter.models import User, Activity


def hash_password(password: str) -> str:
    """Hash a password, return hashed password"""

    if password == '':
        raise ValueError('The password cannot be empty.')

    if len(password) < 8:
        raise ValueError('The password must be at least 8 characters.')

    if len(password) > 24:
        raise ValueError('The password cannot be more than 24 characters.')

    return bcrypt.hashpw(
        password.encode('utf8'), bcrypt.gensalt(rounds=12)
    ).decode('utf8')


def verify_password(password: str, hashed_password: str) -> bool:
    """Verify a password, return true if verified, false if not"""

    return bcrypt.checkpw(
        password.encode('utf8'), hashed_password.encode('utf8')
    )


class UserController(BaseController):
    """User controller encapsulates user management functionality

    Attributes
    ----------
    _instance : UserController
        The instance of the user controller
    _config: ConfigParser
        The configuration parser
    _owner : User
        The current user of the user controller
    _session : Session
        The database session

    Methods
    -------
    create_user(username: str, password: str, email: str)
        Create a new user
    register_user(username: str, password: str, repassword: str, email: str, reemail: str)
        Register a new user identity
    activate_user(user_id: int)
        Activate a user
    deactivate_user(user_id: int)
        Deactivate a user
    login(username: str, password: str)
        User login
    change_password(user_id: int, old_password: str, new_password, repassword: str)
        Change user password
    delete_user(user_id: int)
        Delete a user
    get_user_by_id(user_id: int)
        Get a user by id
    get_user_by_uuid(user_uuid: str)
        Get a user by uuid
    get_user_by_username(username: str)
        Get a user by username
    get_user_by_email(email: str)
        Get a user by email
    get_user_count()
        Get user count
    get_all_users()
        Get all users
    get_all_users_page(page: int, per_page: int)
        Get a single page of users from the database
    search_users(search: str)
        Search for users by username or email
    """

    def __init__(
        self, config: ConfigParser, session: Session, owner: Type[User]
    ):
        """Initialize the class"""

        super().__init__(config, session, owner)

    def create_user(self, username: str, password: str, email: str) -> User:
        """Create a new user

        This method is available to the desktop application owner and the web application owner. It is not available to
        secondary users of the desktop application or the web application. The method first checks if the username or
        email already exists. If either exists, an exception is raised. If neither exists, the method creates a new user
        and logs the activity. The method returns the new user's id on success and nothing on failure.

        Parameters
        ----------
        username : str
            The username of the new user
        password : str
            The password of the new user
        email : str
            The email of the new user

        Returns
        -------
        User
            The new user object
        """

        with self._session as session:

            try:

                username_exists = session.query(User).filter(
                    User.username == username
                ).first()

                if username_exists:
                    raise Exception('That username already exists.')

                email_exists = session.query(User).filter(
                    User.email == email
                ).first()

                if email_exists:
                    raise Exception('That email already exists.')

                uuid4 = str(uuid.uuid4())
                uuid_exists = session.query(User).filter(
                    User.uuid == uuid4
                ).first()

                while uuid_exists:

                    uuid4 = str(uuid.uuid4())
                    uuid_exists = session.query(User).filter(
                        User.uuid == uuid4
                    ).first()

                password = hash_password(password)
                created = datetime.now()
                modified = created

                user = User(
                    uuid=uuid4, username=username, password=password,
                    email=email, created=created, modified=modified,
                    is_active=True
                )

                activity = Activity(
                    user_id=self._owner.id, summary=f'User {user.username} \
                    created by {self._owner.username}', created=datetime.now()
                )

                session.add(user)
                session.add(activity)

            except Exception as e:
                session.rollback()
                raise e

            else:
                session.commit()
                return user

    def register_user(
        self, username: str, password: str, repassword: str, email: str,
        reemail: str
    ) -> User:
        """Register a new user identity

        This method is for self-registration of new users. It is available to secondary users of the website and the
        desktop application. The method first checks if the username or email already exists. If either exists, an
        exception is raised. If neither exists, the method creates a new user and logs the activity. The method returns
        the new user's id on success and nothing on failure.

        Parameters
        ----------
        username : str
            The username of the new user
        password : str
            The password of the new user
        repassword : str
            The password of the new user repeated
        email : str
            The email of the new user
        reemail : str
            The email of the new user repeated

        Returns
        -------
        int
            The id of the new user on success
        """

        with self._session as session:

            try:

                username_exists = session.query(User).filter(User.username == username).first()

                if username_exists:
                    raise Exception('That username already exists.')

                if email != reemail:
                    raise Exception('The email addresses do not match.')

                email_exists = session.query(User).filter(User.email == email).first()

                if email_exists:
                    raise Exception('That email address already exists.')

                if password != repassword:
                    raise Exception('The passwords do not match.')

                uuid4 = str(uuid.uuid4())
                uuid_exists = session.query(User).filter(User.uuid == uuid4).first()

                while uuid_exists:
                    uuid4 = str(uuid.uuid4())
                    uuid_exists = session.query(User).filter(User.uuid == uuid4).first()

                password = hash_password(password)
                created = datetime.now()
                modified = created
                user = User(
                    uuid=uuid4, username=username, password=password,
                    email=email, created=created, modified=modified,
                    is_active=False
                )

                session.add(user)

            except Exception as e:
                session.rollback()
                raise e

            else:
                session.commit()

                activity = Activity(
                    user_id=user.id, summary=f'User {user.username} registered \
                    by {user.username}', created=datetime.now()
                )

                try:
                    session.add(activity)

                except Exception as e:
                    session.rollback()
                    raise e

                else:
                    session.commit()
                    return user

    def activate_user(self, user_id: int) -> Type[User]:
        """Activate a user

        Parameters
        ----------
        user_id: int
            The id of the user to activate

        Returns
        -------
        User
            The activated user object
        """

        with self._session as session:

            user = session.query(User).filter(User.id == user_id).first()

            if not user:
                raise ValueError('User not found.')

            try:
                user.is_active = True
                user.modified = datetime.now()
                activity = Activity(
                    user_id=self._owner.id, summary=f'User {user.username} \
                    activated by {self._owner.username}', created=datetime.now()
                )

                session.add(user)
                session.add(activity)

            except Exception as e:
                session.rollback()
                raise e

            else:
                session.commit()
                return user

    def deactivate_user(self, user_id: int) -> Type[User]:
        """Deactivate a user

        Parameters
        ----------
        user_id: int
            The id of the user to deactivate

        Returns
        -------
        User
            The deactivated user object
        """

        with self._session as session:

            user = session.query(User).filter(User.id == user_id).first()

            if not user:
                raise ValueError('User not found.')

            try:
                user.is_active = False
                user.modified = datetime.now()
                activity = Activity(
                    user_id=self._owner.id, summary=f'User {user.username} \
                    deactivated by {self._owner.username}',
                    created=datetime.now()
                )

                session.add(user)
                session.add(activity)

            except Exception as e:
                session.rollback()
                raise e

            else:
                session.commit()
                return user

    def login(self, username: str, password: str) -> Type[User]:
        """User login

        Parameters
        ----------
        username : str
            The username of the user
        password : str
            The password of the user

        Returns
        -------
        User
            The user object on success
        """

        with self._session as session:

            try:
                candidate = session.query(User).filter(User.username == username).first()

                if not candidate:
                    raise Exception('User not found.')

                if not verify_password(password, candidate.password):
                    raise ValueError('Invalid password.')

                if not candidate.is_active:
                    raise ValueError('User account is not activated.')

                activity = Activity(
                    user_id=candidate.id, summary=f'User {candidate.username} \
                    logged in', created=datetime.now()
                )

                session.add(activity)

            except Exception as e:
                session.rollback()
                raise e

            else:
                session.commit()
                return candidate

    def change_password(
        self, user_id: int, old_password: str, new_password, repassword: str
    ) -> User:
        """Change user password

        Parameters
        ----------
        user_id : int
            The id of the user
        old_password : str
            The old password of the user
        new_password : str
            The new password of the user
        repassword : str
            The new password of the user repeated

        Returns
        -------
        User
            The user object on success
        """

        user = self.get_user_by_id(user_id)

        if not user:
            raise ValueError('User not found.')

        if not verify_password(old_password, user.password):
            raise ValueError('Invalid password.')

        if new_password != repassword:
            raise ValueError('The new passwords do not match.')

        new_password = hash_password(new_password)
        user.password = new_password
        user.modified = str(datetime.now())

        with self._session as session:

            try:
                activity = Activity(
                    user_id=user.id, summary=f'User {user.username} changed \
                    their password', created=datetime.now()
                )

                session.add(user)
                session.add(activity)

            except Exception as e:
                session.rollback()
                raise e

            else:
                session.commit()
                return user

    def delete_user(self, user_id: int) -> bool:
        """Delete a user

        Parameters
        ----------
        user_id : int
            The id of the user to delete

        Returns
        -------
        bool
            True on success
        """

        with self._session as session:

            try:
                user = session.query(User).filter(User.id == user_id).first()

                if not user:
                    raise ValueError('User not found.')

                activity = Activity(
                    user_id=self._owner.id, summary=f'User {user.username} \
                    deleted by {self._owner.username}', created=datetime.now()
                )

                session.delete(user)
                session.add(activity)

            except Exception as e:
                session.rollback()
                raise e

            else:
                session.commit()
                return True

    def get_user_by_id(self, user_id: int) -> Type[User] | None:
        """Get a user by id

        Parameters
        ----------
        user_id : int
            The id of the user to get

        Returns
        -------
        User
            The user
        """

        with self._session as session:
            user = session.query(User).filter(User.id == user_id).first()

            if user:
                return user

            return None

    def get_user_by_uuid(self, user_uuid: str) -> Type[User] | None:
        """Get a user by uuid

        Parameters
        ----------
        user_uuid : str
            The uuid of the user to get

        Returns
        -------
        User
            The user
        """

        with self._session as session:
            user = session.query(User).filter(User.uuid == user_uuid).first()

            if user:
                return user

            return None

    def get_user_by_username(self, username: str) -> Type[User] | None:
        """Get a user by username

        Parameters
        ----------
        username : str
            The username of the user to get

        Returns
        -------
        User
            The user
        """

        with self._session as session:
            user = session.query(User).filter(User.username == username).first()

            if user:
                return user

            return None

    def get_user_by_email(self, email: str) -> Type[User] | None:
        """Get a user by email

        Parameters
        ----------
        email : str
            The email of the user to get

        Returns
        -------
        User
            The user
        """

        with self._session as session:
            user = session.query(User).filter(User.email == email).first()

            if user:
                return user

            return None

    def get_user_count(self) -> int:
        """Get user count

        Returns
        -------
        int
            The number of users
        """

        with self._session as session:
            return session.query(func.count(User.id)).scalar()

    def get_all_users(self) -> List[Type[User]]:
        """Get all users

        Returns
        -------
        list
            A list of users
        """

        with self._session as session:
            return session.query(User).all()

    def get_all_users_page(self, page: int, per_page: int) -> List[Type[User]]:
        """Get a single page of users from the database

        Parameters
        ----------
        page : int
            The page number
        per_page : int
            The number of rows per page

        Returns
        -------
        list
            A list of users
        """

        with self._session as session:
            offset = (page - 1) * per_page
            return session.query(User).offset(offset).limit(per_page).all()

    def search_users(self, search: str) -> List[Type[User]]:
        """Search for users by username or email

        Parameters
        ----------
        search : str
            The search string

        Returns
        -------
        list
            A list of users
        """

        with self._session as session:
            return session.query(User).filter(
                User.username.like(f'%{search}%') | User.email.like(f'%{search}%')
            ).all()



from configparser import ConfigParser
from datetime import datetime
from typing import Type, List
from sqlalchemy import or_
from sqlalchemy.orm import Session
from gnatwriter.controllers.BaseController import BaseController
from gnatwriter.models import User, Story, Author, AuthorStory, Chapter, Activity, Character, CharacterStory, Link, \
    LinkStory, Note, NoteStory


class StoryController(BaseController):
    """Story controller encapsulates story management functionality

    Attributes
    ----------
    _instance : StoryController
        The instance of the story controller
    _config: ConfigParser
        The configuration parser
    _owner : User
        The current user of the story controller
    _session : Session
        The database session

    Methods
    -------
    create_story(title: str, description: str)
        Create a new story
    update_story(story_id: int, title: str, description: str)
        Update a story
    delete_story(story_id: int)
        Delete a story
    has_stories() : bool
        Check if a user has stories
    count_stories()
        Count the number of stories associated with a user
    get_story_by_id(story_id: int)
        Get a story by id
    get_all_stories()
        Get all stories associated with an owner
    get_all_stories_page(page: int, per_page: int)
        Get a single page of stories associated with an owner from the database
    search_stories(search: str)
        Search for stories by title and description
    append_authors_to_story(story_id: int, author_ids: list)
        Append authors to a story
    detach_authors_from_story(story_id:int, author_ids: list)
        Detach authors from a story
    has_authors( story_id ) : bool
        Check if a story has authors
    get_authors_by_story_id(story_id: int)
        Get all authors associated with a story
    has_chapters( story_id ) : bool
        Check if a story has chapters
    get_chapter_by_position(story_id: int, position: int)
        Get a chapter by position
    get_all_chapters_by_story_id(story_id: int)
        Get all chapters associated with a story
    append_characters_to_story(story_id: int, character_ids: list)
        Append characters to a story
    has_characters( story_id ) : bool
        Check if a story has characters
    get_characters_by_story_id(story_id: int)
        Get all characters associated with a story
    get_characters_page_by_story_id(story_id: int, page: int, per_page: int)
        Get a single page of characters associated with a story from the database
    append_links_to_story(story_id: int, link_ids: list)
        Append links to a story
    has_links( story_id ) : bool
        Check if a story has links
    get_links_by_story_id(story_id: int)
        Get all links associated with a story
    append_notes_to_story(story_id: int, note_ids: list)
        Append notes to a story
    has_notes( story_id ) : bool
        Check if a story has notes
    get_notes_by_story_id(story_id: int)
        Get all notes associated with a story
    """

    def __init__(
        self, config: ConfigParser, session: Session, owner: Type[User]
    ):
        """Initialize the class"""

        super().__init__(config, session, owner)

    def create_story(self, title: str, description: str = None) -> Story:
        """Create a new story

        Parameters
        ----------
        title : str
            The title of the story
        description : str
            The description of the story

        Returns
        -------
        Story
            The new story object
        """

        with self._session as session:
            try:
                created = datetime.now()
                modified = created

                story = Story(
                    user_id=self._owner.id, title=title, description=description,
                    created=created, modified=modified
                )

                activity = Activity(
                    user_id=self._owner.id, summary=f'Story {story.title[:50]} \
                    created by {self._owner.username}', created=datetime.now()
                )

                session.add(story)
                session.add(activity)

            except Exception as e:
                session.rollback()
                raise e

            else:
                session.commit()
                return story

    def update_story(
        self, story_id: int, title: str, description: str = None
    ) -> Type[Story]:
        """Update a story

        Parameters
        ----------
        story_id : int
            The id of the story
        title : str
            The title of the story
        description : str
            The description of the story

        Returns
        -------
        bool
            True on success
        """

        with self._session as session:
            try:
                story = session.query(Story).filter(
                    Story.id == story_id,
                    Story.user_id == self._owner.id
                ).first()

                if not story:
                    raise ValueError('Story not found.')

                story.title = title
                story.description = description
                story.modified = datetime.now()

                activity = Activity(
                    user_id=self._owner.id, summary=f'Story {story.id} updated \
                    by {self._owner.username}', created=datetime.now()
                )

                session.add(activity)

            except Exception as e:
                session.rollback()
                raise e

            else:
                session.commit()
                return story

    def delete_story(self, story_id: int) -> bool:
        """Delete a story

        Parameters
        ----------
        story_id : int
            The id of the story

        Returns
        -------
        bool
            True on success
        """

        with self._session as session:
            try:
                story = session.query(Story).filter(
                    Story.id == story_id,
                    Story.user_id == self._owner.id
                ).first()

                if not story:
                    raise ValueError('Story not found.')

                activity = Activity(
                    user_id=self._owner.id, summary=f'Story {story.id} deleted \
                    by {self._owner.username}', created=datetime.now()
                )

                session.delete(story)
                session.add(activity)

            except Exception as e:
                session.rollback()
                raise e

            else:
                session.commit()
                return True

    def has_stories(self) -> bool:
        """Check if a user has stories

        Returns
        -------
        bool
            True if the user has stories
        """

        with self._session as session:

            return session.query(Story).filter(
                Story.user_id == self._owner.id
            ).count() > 0

    def count_stories(self) -> int:
        """Count the number of stories associated with a user

        Returns
        -------
        int
            The number of stories
        """

        with self._session as session:

            return session.query(Story).filter(
                Story.user_id == self._owner.id
            ).count()

    def get_story_by_id(self, story_id: int) -> Type[Story] | None:
        """Get a story by id

        Parameters
        ----------
        story_id : int
            The id of the story

        Returns
        -------
        Story
            The story object
        """

        with self._session as session:
            story = session.query(Story).filter(
                Story.id == story_id,
                Story.user_id == self._owner.id
            ).first()
            return story

    def get_all_stories(self) -> List[Type[Story]]:
        """Get all stories associated with an owner

        Returns
        -------
        list
            A list of story objects
        """

        with self._session as session:
            return session.query(Story).filter(
                Story.user_id == self._owner.id
            ).all()

    def get_all_stories_page(
        self, page: int, per_page: int
    ) -> List[Type[Story]]:
        """Get a single page of stories associated with an owner from the database

        Parameters
        ----------
        page : int
            The page number
        per_page : int
            The number of rows per page

        Returns
        -------
        list
            A list of story objects
        """

        with self._session as session:
            offset = (page - 1) * per_page
            return session.query(Story).filter(
                Story.user_id == self._owner.id
            ).offset(offset).limit(per_page).all()

    def search_stories(self, search: str) -> List[Type[Story]]:
        """Search for stories by title and description

        Parameters
        ----------
        search : str
            The search string

        Returns
        -------
        list
            A list of story objects
        """

        with self._session as session:
            return session.query(Story).filter(
                or_(
                    Story.title.like(f'%{search}%'),
                    Story.description.like(f'%{search}%')
                ),
                Story.user_id == self._owner.id
            ).all()

    def append_authors_to_story(
        self, story_id: int, author_ids: list
    ) -> Type[Story]:
        """Append authors to a story

        Parameters
        ----------
        story_id : int
            The id of the story
        author_ids : list
            The ids of the authors to append

        Returns
        -------
        Story
            The updated story object
        """

        with self._session as session:
            try:
                story = session.query(Story).filter(
                    Story.id == story_id,
                    Story.user_id == self._owner.id
                ).first()

                if not story:
                    raise ValueError('Story not found.')

                for author_id in author_ids:
                    author = session.query(Author).filter(
                        Author.id == author_id,
                        Author.user_id == self._owner.id
                    ).first()

                    if not author:
                        raise ValueError('Author not found.')

                    author_story = session.query(AuthorStory).filter(
                        AuthorStory.user_id == self._owner.id,
                        AuthorStory.author_id == author_id,
                        AuthorStory.story_id == story_id
                    ).first()

                    if not author_story:

                        author_story = AuthorStory(
                            user_id=self._owner.id, author_id=author_id,
                            story_id=story_id, created=datetime.now()
                        )
                        story.authors.append(author_story)

                        activity = Activity(
                            user_id=self._owner.id, summary=f'Authors appended to \
                            story {story.title[:50]} by {self._owner.username}',
                            created=datetime.now()
                        )

                        session.add(activity)

            except Exception as e:
                session.rollback()
                raise e

            else:
                session.commit()
                return story

    def detach_authors_from_story(
        self, story_id: int, author_ids: list
    ) -> Type[Story]:
        """Detach authors from a story

        Parameters
        ----------
        story_id : int
            The id of the story
        author_ids : list
            The ids of the authors to detach

        Returns
        -------
        Story
            The updated story object
        """

        with self._session as session:
            try:
                story = session.query(Story).filter(
                    Story.id == story_id,
                    Story.user_id == self._owner.id
                ).first()

                if not story:
                    raise ValueError('Story not found.')

                for author_id in author_ids:
                    author = session.query(Author).filter(
                        Author.id == author_id,
                        Author.user_id == self._owner.id
                    ).first()

                    if not author:
                        raise ValueError('Author not found.')

                    author_story = session.query(AuthorStory).filter(
                        AuthorStory.user_id == self._owner.id,
                        AuthorStory.author_id == author_id,
                        AuthorStory.story_id == story_id
                    ).first()

                    if not author_story:
                        return story

                    activity = Activity(
                        user_id=self._owner.id, summary=f'Authors detached from \
                        story {story.title[:50]} by {self._owner.username}',
                        created=datetime.now()
                    )

                    session.delete(author_story)
                    session.add(activity)

            except Exception as e:
                session.rollback()
                raise e

            else:
                session.commit()
                return story

    def has_authors(self, story_id: int) -> bool:
        """Check if a story has authors

        Parameters
        ----------
        story_id : int
            The id of the story

        Returns
        -------
        bool
            True if the story has authors
        """

        with self._session as session:

            return session.query(Author).join(AuthorStory).filter(
                AuthorStory.story_id == story_id,
                AuthorStory.user_id == self._owner.id
            ).count() > 0

    def get_authors_by_story_id(self, story_id: int) -> List[Type[Author]]:
        """Get all authors associated with a story

        Parameters
        ----------
        story_id : int
            The id of the story

        Returns
        -------
        list
            A list of author objects
        """

        with self._session as session:
            return session.query(Author).join(AuthorStory).filter(
                AuthorStory.story_id == story_id,
                AuthorStory.user_id == self._owner.id
            ).all()

    def has_chapters(self, story_id: int) -> bool:
        """Check if a story has chapters

        Parameters
        ----------
        story_id : int
            The id of the story

        Returns
        -------
        bool
            True if the story has chapters
        """

        with self._session as session:

            return session.query(Chapter).filter(
                Chapter.story_id == story_id,
                Chapter.user_id == self._owner.id
            ).count() > 0

    def get_chapter_by_position(
        self, story_id: int, position: int
    ) -> Type[Chapter]:
        """Get a chapter by position

        Parameters
        ----------
        story_id : int
            The id of the story
        position : int
            The position of the chapter

        Returns
        -------
        Chapter
            The chapter object
        """

        with self._session as session:

            return session.query(Chapter).filter(
                Chapter.story_id == story_id,
                Chapter.user_id == self._owner.id,
                Chapter.position == position
            ).first()

    def get_all_chapters_by_story_id(
        self, story_id: int
    ) -> List[Type[Chapter]]:
        """Get all chapters associated with a story

        Parameters
        ----------
        story_id : int
            The id of the story

        Returns
        -------
        list
            A list of chapter objects
        """

        with self._session as session:

            return session.query(Chapter).filter(
                Chapter.story_id == story_id,
                Chapter.user_id == self._owner.id
            ).order_by(Chapter.position).all()

    def append_characters_to_story(
        self, story_id: int, character_ids: list
    ) -> Type[Story]:
        """Append characters to a story

        Parameters
        ----------
        story_id : int
            The id of the story
        character_ids : list
            The ids of the characters to append

        Returns
        -------
        Story
            The updated story object
        """

        with self._session as session:

            try:

                story = session.query(Story).filter(
                    Story.id == story_id,
                    Story.user_id == self._owner.id
                ).first()

                if not story:
                    raise ValueError('Story not found.')

                for character_id in character_ids:
                    character = session.query(Character).filter(
                        Character.id == character_id,
                        Character.user_id == self._owner.id
                    ).first()

                    if not character:
                        raise ValueError('Character not found.')

                    character_story = CharacterStory(
                        user_id=self._owner.id, character_id=character_id,
                        story_id=story_id, created=datetime.now()
                    )

                    story.characters.append(character_story)

                activity = Activity(
                    user_id=self._owner.id, summary=f'Characters appended to \
                    story {story.title[:50]} by {self._owner.username}',
                    created=datetime.now()
                )

                session.add(activity)

            except Exception as e:
                session.rollback()
                raise e

            else:
                session.commit()
                return story

    def has_characters(self, story_id: int) -> bool:
        """Check if a story has characters

        Parameters
        ----------
        story_id : int
            The id of the story

        Returns
        -------
        bool
            True if the story has characters
        """

        with self._session as session:

            return session.query(Character).join(CharacterStory).filter(
                CharacterStory.story_id == story_id, Character
            ).count() > 0

    def get_characters_by_story_id(
        self, story_id: int
    ) -> List[Type[Character]]:
        """Get all characters associated with a story

        Parameters
        ----------
        story_id : int
            The id of the story

        Returns
        -------
        list
            A list of character objects
        """

        with self._session as session:

            return session.query(Character).join(CharacterStory).filter(
                CharacterStory.story_id == story_id, Character
            ).all()

    def get_characters_page_by_story_id(
        self, story_id: int, page: int, per_page: int
    ) -> List[Type[Character]]:
        """Get a single page of characters associated with a story from the database

        Parameters
        ----------
        story_id : int
            The id of the story
        page : int
            The page number
        per_page : int
            The number of rows per page

        Returns
        -------
        list
            A list of character objects
        """

        with self._session as session:

            offset = (page - 1) * per_page

            return session.query(Character).join(CharacterStory).filter(
                CharacterStory.story_id == story_id, Character
            ).offset(offset).limit(per_page).all()

    def append_links_to_story(
        self, story_id: int, link_ids: list
    ) -> Type[Story]:
        """Append links to a story

        Parameters
        ----------
        story_id : int
            The id of the story
        link_ids : list
            The ids of the links to append

        Returns
        -------
        Story
            The updated story object
        """

        with self._session as session:
            try:
                story = session.query(Story).filter(
                    Story.id == story_id,
                    Story.user_id == self._owner.id
                ).first()

                if not story:
                    raise ValueError('Story not found.')

                for link_id in link_ids:
                    link = session.query(Link).filter(
                        Link.id == link_id,
                        Link.user_id == self._owner.id
                    ).first()

                    if not link:
                        raise ValueError('Link not found.')

                    link_story = LinkStory(
                        user_id=self._owner.id, story_id=story_id,
                        link_id=link_id, created=datetime.now()
                    )

                    story.links.append(link_story)

                activity = Activity(
                    user_id=self._owner.id, summary=f'Links appended to story \
                    {story.title[:50]} by {self._owner.username}',
                    created=datetime.now()
                )

                session.add(activity)

            except Exception as e:
                session.rollback()
                raise e

            else:
                session.commit()
                return story

    def has_links(self, story_id: int) -> bool:
        """Check if a story has links

        Parameters
        ----------
        story_id : int
            The id of the story

        Returns
        -------
        bool
            True if the story has links
        """

        with self._session as session:
            return session.query(Link).join(LinkStory).filter(
                LinkStory.story_id == story_id, LinkStory
            ).count() > 0

    def get_links_by_story_id(self, story_id: int) -> List[Type[Link]]:
        """Get all links associated with a story

        Parameters
        ----------
        story_id : int
            The id of the story

        Returns
        -------
        list
            A list of link objects
        """

        with self._session as session:
            return session.query(Link).join(LinkStory).filter(
                LinkStory.story_id == story_id,
                LinkStory.user_id == self._owner.id
            ).all()

    def append_notes_to_story(
        self, story_id: int, note_ids: list
    ) -> Type[Story]:
        """Append notes to a story

        Parameters
        ----------
        story_id : int
            The id of the story
        note_ids : list
            The ids of the notes to append

        Returns
        -------
        Story
            The updated story object
        """

        with self._session as session:
            try:
                story = session.query(Story).filter(
                    Story.id == story_id,
                    Story.user_id == self._owner.id
                ).first()

                if not story:
                    raise ValueError('Story not found.')

                for note_id in note_ids:
                    note = session.query(Note).filter(
                        Note.id == note_id,
                        Note.user_id == self._owner.id
                    ).first()

                    if not note:
                        raise ValueError('Note not found.')

                    note_story = NoteStory(
                        user_id=self._owner.id, story_id=story_id,
                        note_id=note_id, created=datetime.now()
                    )

                    story.notes.append(note_story)

                activity = Activity(
                    user_id=self._owner.id, summary=f'Notes appended to story \
                    {story.title[:50]} by {self._owner.username}',
                    created=datetime.now()
                )

                session.add(activity)

            except Exception as e:
                session.rollback()
                raise e

            else:
                session.commit()
                return story

    def has_notes(self, story_id: int) -> bool:
        """Check if a story has notes

        Parameters
        ----------
        story_id : int
            The id of the story

        Returns
        -------
        bool
            True if the story has notes
        """

        with self._session as session:

            return session.query(Note).join(NoteStory).filter(
                NoteStory.story_id == story_id, NoteStory
            ).count() > 0

    def get_notes_by_story_id(self, story_id: int) -> List[Type[Note]]:
        """Get all notes associated with a story

        Parameters
        ----------
        story_id : int
            The id of the story

        Returns
        -------
        list
            A list of note objects
        """

        with self._session as session:
            return session.query(Note).join(NoteStory).filter(
                NoteStory.story_id == story_id,
                NoteStory.user_id == self._owner.id
            ).all()



from configparser import ConfigParser
from datetime import datetime
from typing import Type, List
from sqlalchemy.orm import Session
from gnatwriter.controllers.BaseController import BaseController
from gnatwriter.models import User, Image, Activity, ImageLocation


class ImageController(BaseController):
    """Image controller encapsulates image management functionality

    Attributes
    ----------
    _instance : ImageController
        The instance of the image controller
    _config: ConfigParser
        The configuration parser
    _owner : User
        The current user of the image controller
    _session : Session
        The database session

    Methods
    -------
    create_image(caption: str, filename: str, dirname: str, size_in_bytes: int, mime_type: str)
        Create a new image
    update_image(image_id: int, caption: str)
        Update an image
    delete_image(image_id: int)
        Delete an image
    get_all_images()
        Get all images associated with a user
    get_all_images_page(page: int, per_page: int)
        Get a single page of images associated with a user from the database
    search_images(search: str)
        Search for images by caption
    get_images_by_character_id(character_id: int)
        Get all images associated with a character
    get_images_page_by_character_id(character_id: int, page: int, per_page: int)
        Get a single page of images associated with a character from the database
    get_images_by_location_id(location_id: int)
        Get all images associated with a location
    get_images_page_by_location_id(location_id: int, page: int, per_page: int)
        Get a single page of images associated with a location from the database
    """

    def __init__(
        self, config: ConfigParser, session: Session, owner: Type[User]
    ):
        """Initialize the class"""

        super().__init__(config, session, owner)

    def create_image(
        self, filename: str, dirname: str, size_in_bytes: int, mime_type: str,
        caption: str = None
    ) -> Image:
        """Create a new image

        Parameters
        ----------
        filename : str
            The filename of the image
        dirname : str
            The directory name of the image
        size_in_bytes : int
            The size of the image in bytes
        mime_type : str
            The mime type of the image
        caption : str
            The caption of the image, optional

        Returns
        -------
        Image
            The new image object
        """

        with self._session as session:
            try:
                created = datetime.now()
                modified = created

                image = Image(
                    user_id=self._owner.id, caption=caption, filename=filename,
                    dirname=dirname, size_in_bytes=size_in_bytes,
                    mime_type=mime_type, created=created, modified=modified
                )

                activity = Activity(
                    user_id=self._owner.id, summary=f'Image {image.id} created \
                    by {self._owner.username}', created=datetime.now()
                )

                session.add(image)
                session.add(activity)

            except Exception as e:
                session.rollback()
                raise e

            else:
                session.commit()
                return image

    def update_image(self, image_id: int, caption: str = None) -> Type[Image]:
        """Update an image

        Parameters
        ----------
        image_id : int
            The id of the image
        caption : str
            The caption of the image, optional

        Returns
        -------
        Image
            The updated image object
        """

        with self._session as session:
            try:
                image = session.query(Image).filter(
                    Image.id == image_id,
                    Image.user_id == self._owner.id
                ).first()

                if not image:
                    raise ValueError('Image not found.')

                image.caption = caption
                image.modified = datetime.now()

                activity = Activity(
                    user_id=self._owner.id, summary=f'Image {image.id} updated \
                    by {self._owner.username}', created=datetime.now()
                )

                session.add(activity)

            except Exception as e:
                session.rollback()
                raise e

            else:
                session.commit()
                return image

    def delete_image(self, image_id: int) -> bool:
        """Delete an image

        Parameters
        ----------
        image_id : int
            The id of the image

        Returns
        -------
        bool
            True on success
        """

        with self._session as session:
            try:
                image = session.query(Image).filter(
                    Image.id == image_id,
                    Image.user_id == self._owner.id
                ).first()

                if not image:
                    raise ValueError('Image not found.')

                activity = Activity(
                    user_id=self._owner.id, summary=f'Image {image.id} deleted \
                    by {self._owner.username}', created=datetime.now()
                )

                session.delete(image)
                session.add(activity)

            except Exception as e:
                session.rollback()
                raise e

            else:
                session.commit()
                return True

    def get_all_images(self) -> List[Type[Image]]:
        """Get all images associated with a user

        Returns
        -------
        list
            A list of image objects
        """

        with self._session as session:
            return session.query(Image).filter(
                Image.user_id == self._owner.id
            ).all()

    def get_all_images_page(
        self, page: int, per_page: int
    ) -> List[Type[Image]]:
        """Get a single page of images associated with a user from the database

        Parameters
        ----------
        page : int
            The page number
        per_page : int
            The number of rows per page

        Returns
        -------
        list
            A list of image objects
        """

        with self._session as session:
            offset = (page - 1) * per_page
            return session.query(Image).filter(
                Image.user_id == self._owner.id
            ).offset(offset).limit(per_page).all()

    def search_images(self, search: str) -> List[Type[Image]]:
        """Search for images by caption

        Parameters
        ----------
        search : str
            The search string

        Returns
        -------
        list
            A list of image objects
        """

        with self._session as session:
            return session.query(Image).filter(
                Image.caption.like(f'%{search}%'),
                Image.user_id == self._owner.id
            ).all()

    def get_images_by_character_id(
        self, character_id: int
    ) -> List[Type[Image]]:
        """Get all images associated with a character

        Parameters
        ----------
        character_id : int
            The id of the character

        Returns
        -------
        list
            A list of image objects
        """

        with self._session as session:
            return session.query(Image).filter(
                Image.character_id == character_id,
                Image.user_id == self._owner.id
            ).all()

    def get_images_page_by_character_id(
        self, character_id: int, page: int, per_page: int
    ) -> List[Type[Image]]:
        """Get a single page of images associated with a character from the database

        Parameters
        ----------
        character_id : int
            The id of the character
        page : int
            The page number
        per_page : int
            The number of rows per page

        Returns
        -------
        list
            A list of image objects
        """

        with self._session as session:
            offset = (page - 1) * per_page
            return session.query(Image).filter(
                Image.character_id == character_id,
                Image.user_id == self._owner.id
            ).offset(offset).limit(per_page).all()

    def get_images_by_location_id(self, location_id: int) -> List[Type[Image]]:
        """Get all images associated with a location

        Images and Locations are associated through ImageLocation objects. This method will use yield to return the
        images one at a time.

        Parameters
        ----------
        location_id : int
            The id of the location

        Returns
        -------
        list
            A list of image objects
        """

        with self._session as session:
            for image_location in session.query(ImageLocation).filter(
                ImageLocation.location_id == location_id,
                    ImageLocation.user_id == self._owner.id
            ).all():
                yield session.query(Image).filter(
                    Image.id == image_location.image_id,
                    Image.user_id == self._owner.id
                ).first()

    def get_images_page_by_location_id(
        self, location_id: int, page: int, per_page: int
    ) -> List[Type[Image]]:
        """Get a single page of images associated with a location from the database

        Images and Locations are associated through ImageLocation objects. This method will use yield to return the
        images one at a time.

        Parameters
        ----------
        location_id : int
            The id of the location
        page : int
            The page number
        per_page : int
            The number of rows per page

        Returns
        -------
        list
            A list of image objects
        """

        with self._session as session:
            offset = (page - 1) * per_page
            for image_location in session.query(ImageLocation).filter(
                ImageLocation.location_id == location_id,
                ImageLocation.user_id == self._owner.id
            ).offset(offset).limit(per_page).all():
                yield session.query(Image).filter(
                    Image.id == image_location.image_id,
                    Image.user_id == self._owner.id
                ).first()



from datetime import datetime
from sqlalchemy import Integer, String, Text, DateTime
from sqlalchemy.orm import Mapped, mapped_column, validates
from gnatwriter.models import Base


class OllamaModel(Base):
    """OllamaModel stores the name of the model, a description of the model, the template, and an example of the
    model's use.
    """

    __tablename__ = "ollama_models"
    id: Mapped[int] = mapped_column(Integer, primary_key=True)
    title: Mapped[str] = mapped_column(String(150), nullable=False)
    model: Mapped[str] = mapped_column(String(75), nullable=False)
    description: Mapped[str] = mapped_column(Text, nullable=True)
    template: Mapped[str] = mapped_column(Text, nullable=False)
    example: Mapped[str] = mapped_column(Text, nullable=True)
    priming: Mapped[str] = mapped_column(Text, nullable=True)
    params: Mapped[str] = mapped_column(Text, nullable=True)
    created: Mapped[str] = mapped_column(DateTime, default=str(datetime.now()))
    modified: Mapped[str] = mapped_column(
        DateTime, default=str(datetime.now()), onupdate=str(datetime.now())
    )

    def __repr__(self) -> str:
        return f"<OllamaModel(title={self.title}, model={self.model})>"

    def __str__(self) -> str:
        return f"{self.title} ({self.model})"

    def serialize(self) -> dict:
        return {
            "id": self.id,
            "title": self.title,
            "model": self.model,
            "description": self.description,
            "template": self.template,
            "example": self.example,
            "priming": self.priming,
            "params": self.params,
            "created": str(self.created),
            "modified": str(self.modified),
        }

    def unserialize(self, data: dict) -> "OllamaModel":
        self.title = data.get("title", self.title)
        self.model = data.get("model", self.model)
        self.description = data.get("description", self.description)
        self.template = data.get("template", self.template)
        self.example = data.get("example", self.example)
        self.priming = data.get("priming", self.priming)
        self.params = data.get("params", self.params)
        self.created = data.get("created", self.created)
        self.modified = data.get("modified", self.modified)

        return self

    @validates("title")
    def validate_title(self, key, title: str) -> str:
        """Validates the title's length.

        Parameters
        ----------
        title: str
            The character's title

        Returns
        -------
        str
            The validated title
        """

        if title and len(title) > 150:
            raise ValueError("The model's title must have no more than 150 characters.")

        return title

    @validates("model")
    def validate_model(self, key, model: str) -> str:
        """Validates the model's length.

        Parameters
        ----------
        model: str
            The character's model

        Returns
        -------
        str
            The validated model
        """

        if model and len(model) > 75:
            raise ValueError("The model's name must have no more than 75 characters.")

        return model

    @validates("description")
    def validate_description(self, key, description: str) -> str:
        """Validates the description's length.

        Parameters
        ----------
        description: str
            The character's description

        Returns
        -------
        str
            The validated description
        """

        if description and len(description) > 65535:
            raise ValueError("The model's description must have no more than 65,535 characters.")

        return description

    @validates("template")
    def validate_template(self, key, template: str) -> str:
        """Validates the template's length.

        Parameters
        ----------
        template: str
            The character's template

        Returns
        -------
        str
            The validated template
        """

        if template and len(template) > 65535:
            raise ValueError("The model's template must have no more than 65,535 characters.")

        return template

    @validates("example")
    def validate_example(self, key, example: str) -> str:
        """Validates the example's length.

        Parameters
        ----------
        example: str
            The character's example

        Returns
        -------
        str
            The validated example
        """

        if example and len(example) > 65535:
            raise ValueError("The model's example must have no more than 65,535 characters.")

        return example

    @validates("priming")
    def validate_priming(self, key, priming: str) -> str:
        """Validates the priming's length.

        Parameters
        ----------
        priming: str
            The character's priming

        Returns
        -------
        str
            The validated priming
        """

        if priming and len(priming) > 65535:
            raise ValueError("The model's priming must have no more than 65,535 characters.")

        return priming

    @validates("params")
    def validate_params(self, key, params: str) -> str:
        """Validates the params' length.

        Parameters
        ----------
        params: str
            The character's params

        Returns
        -------
        str
            The validated params
        """

        if params and len(params) > 65535:
            raise ValueError("The model's params must have no more than 65,535 characters.")

        return params
